'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var node_console = require('node:console');
var path = require('node:path');
var node_url = require('node:url');
var puppeteer = require('puppeteer');
var childProcess = require('node:child_process');
var node_fs = require('node:fs');
var http = require('http');
var querystring = require('querystring');
var postcssPlugin = require('rollup-plugin-postcss');
var acorn = require('acorn');
var esModuleLexer = require('es-module-lexer');
var mime = require('mime/lite');
var esbuild = require('esbuild');
var node_crypto = require('node:crypto');
var node_module = require('node:module');
var commonjs = require('@rollup/plugin-commonjs');
var cjsModuleLexer = require('cjs-module-lexer');
var rollup = require('rollup');

function _interopDefault (e) { return e && e.__esModule ? e["default"] : e; }

function _interopNamespace(e) {
	if (e && e.__esModule) return e;
	var n = Object.create(null);
	if (e) {
		for (var k in e) {
			n[k] = e[k];
		}
	}
	n["default"] = e;
	return Object.freeze(n);
}

var path__namespace = /*#__PURE__*/_interopNamespace(path);
var path__default = /*#__PURE__*/_interopDefault(path);
var childProcess__namespace = /*#__PURE__*/_interopNamespace(childProcess);
var http__namespace = /*#__PURE__*/_interopNamespace(http);
var postcssPlugin__default = /*#__PURE__*/_interopDefault(postcssPlugin);
var mime__default = /*#__PURE__*/_interopDefault(mime);
var esbuild__namespace = /*#__PURE__*/_interopNamespace(esbuild);
var commonjs__default = /*#__PURE__*/_interopDefault(commonjs);

function ansiRegex$1({
  onlyFirst = false
} = {}) {
  const pattern = ['[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]+)*|[a-zA-Z\\d]+(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]*)*)?\\u0007)', '(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PR-TZcf-ntqry=><~]))'].join('|');
  return new RegExp(pattern, onlyFirst ? undefined : 'g');
}

function createRawFrame(raw) {
  return {
    column: -1,
    fileName: "",
    line: -1,
    name: "",
    raw: raw,
    sourceColumn: -1,
    sourceFileName: "",
    sourceLine: -1,
    type: ""
  };
}

var FIREFOX_WEBKIT = /([^@]+|^)@(.*):(\d+):(\d+)/;
var WEBKIT_ADDRESS_UNNAMED = /^(http(s)?:\/\/.*):(\d+):(\d+)$/;
var WEBKIT_NATIVE_UNNAMED = "[native code]";

function parsWebkit(str) {
  if (str.includes(WEBKIT_NATIVE_UNNAMED)) {
    return {
      line: -1,
      column: -1,
      type: "native",
      fileName: "",
      name: "",
      raw: str,
      sourceColumn: -1,
      sourceFileName: "",
      sourceLine: -1
    };
  }

  var match = str.match(WEBKIT_ADDRESS_UNNAMED);

  if (match) {
    var line = match[3] ? +match[3] : -1;
    var column = match[4] ? +match[4] : -1;
    var fileName = match[1] ? match[1] : "";
    return {
      line: line,
      column: column,
      type: "",
      fileName: fileName,
      name: "",
      raw: str,
      sourceColumn: -1,
      sourceFileName: "",
      sourceLine: -1
    };
  }

  return createRawFrame(str);
}

function parseFirefoxWebkit(lines) {
  return lines.map(function (str) {
    var match = str.match(FIREFOX_WEBKIT);

    if (!match) {
      return parsWebkit(str);
    }

    var line = match[3] ? +match[3] : -1;
    var column = match[4] ? +match[4] : -1;
    var fileName = match[2] ? match[2] : "";
    return {
      line: line,
      column: column,
      type: match[0] ? "" : "native",
      fileName: fileName,
      name: (match[1] || "").trim(),
      raw: str,
      sourceColumn: -1,
      sourceFileName: "",
      sourceLine: -1
    };
  });
} // foo.bar.js:123:39
// foo.bar.js:123:39 <- original.js:123:34


var CHROME_MAPPED = /(.*?):(\d+):(\d+)(\s<-\s(.+):(\d+):(\d+))?/;

function parseMapped(frame, maybeMapped) {
  var match = maybeMapped.match(CHROME_MAPPED);

  if (match) {
    frame.fileName = match[1];
    frame.line = +match[2];
    frame.column = +match[3];
    if (match[5]) frame.sourceFileName = match[5];
    if (match[6]) frame.sourceLine = +match[6];
    if (match[7]) frame.sourceColumn = +match[7];
  }
} // at <SomeFramework>


var CHROME_IE_NATIVE_NO_LINE = /^at\s(<.*>)$/; // at <SomeFramework>:123:39

var CHROME_IE_NATIVE = /^\s*at\s(<.*>):(\d+):(\d+)$/; // at foo.bar(bob) (foo.bar.js:123:39)
// at foo.bar(bob) (foo.bar.js:123:39 <- original.js:123:34)

var CHROME_IE_FUNCTION = /^at\s(.*)\s\((.*)\)$/; // >= Chrome 88
// spy() at Component.Foo [as constructor] (original.js:123:34)
// spy() at Component.Foo [as constructor] (foo.bar.js:123:39 <- original.js:123:34)

var CHROME_IE_FUNCTION_WITH_CALL = /^([\w\W]*)\s\((.*)\)/;
var CHROME_IE_DETECTOR = /\s*at\s.+/; // at about:blank:1:7
// >= Chrome 99
// at /projects/foo.test.js:689:1 <- /projects/foo.test.js:10:1

var CHROME_BLANK = /\s*at\s(.*):(\d+):(\d+)$/;

function parseChromeIe(lines) {
  // Many frameworks mess with error.stack. So we use this check
  // to find the first line of the actual stack
  var start = lines.findIndex(function (line) {
    return CHROME_IE_DETECTOR.test(line);
  });

  if (start === -1) {
    return [];
  }

  var frames = [];

  for (var i = start; i < lines.length; i++) {
    var str = lines[i].replace(/^\s+|\s+$/g, "");
    var frame = createRawFrame(lines[i]);
    var nativeNoLine = str.match(CHROME_IE_NATIVE_NO_LINE);

    if (nativeNoLine) {
      frame.fileName = nativeNoLine[1];
      frame.type = "native";
      frames.push(frame);
      continue;
    }

    var native = str.match(CHROME_IE_NATIVE);

    if (native) {
      frame.fileName = native[1];
      frame.type = "native";
      if (native[2]) frame.line = +native[2];
      if (native[3]) frame.column = +native[3];
      frames.push(frame);
      continue;
    }

    var withFn = str.match(CHROME_IE_FUNCTION);

    if (withFn) {
      frame.name = withFn[1];
      parseMapped(frame, withFn[2]);
      frames.push(frame);
      continue;
    }

    var blank = str.match(CHROME_BLANK);

    if (blank) {
      frame.fileName = blank[1];
      frame.line = +blank[2];
      frame.column = +blank[3];
      parseMapped(frame, blank[1] + ":" + blank[2] + ":" + blank[3]);
      frames.push(frame);
      continue;
    }

    var withFnCall = str.match(CHROME_IE_FUNCTION_WITH_CALL);

    if (withFnCall) {
      frame.name = withFnCall[1];
      parseMapped(frame, withFnCall[2]);
      frames.push(frame);
      continue;
    }

    frames.push(frame);
  }

  return frames;
}

function parseStackTrace(stack) {
  if (!stack) return [];
  var lines = stack.split("\n").filter(Boolean); // Libraries like node's "assert" module mess with the stack trace by
  // prepending custom data. So we need to do a precheck, to determine
  // which browser the trace is coming from.

  if (lines.some(function (line) {
    return CHROME_IE_DETECTOR.test(line);
  })) {
    return parseChromeIe(lines);
  }

  return parseFirefoxWebkit(lines);
}

let enabled = true; // Support both browser and node environments

const globalVar = typeof self !== 'undefined' ? self : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : {};
/**
 * Detect how much colors the current terminal supports
 */

let supportLevel = 0
/* none */
;

if (globalVar.process && globalVar.process.env && globalVar.process.stdout) {
  const {
    FORCE_COLOR,
    NODE_DISABLE_COLORS,
    TERM
  } = globalVar.process.env;

  if (NODE_DISABLE_COLORS || FORCE_COLOR === '0') {
    enabled = false;
  } else if (FORCE_COLOR === '1') {
    enabled = true;
  } else if (TERM === 'dumb') {
    enabled = false;
  } else if ('CI' in globalVar.process.env && ['TRAVIS', 'CIRCLECI', 'APPVEYOR', 'GITLAB_CI', 'GITHUB_ACTIONS', 'BUILDKITE', 'DRONE'].some(vendor => vendor in globalVar.process.env)) {
    enabled = true;
  } else {
    enabled = process.stdout.isTTY;
  }

  if (enabled) {
    supportLevel = TERM && TERM.endsWith('-256color') ? 2
    /* ansi256 */
    : 1
    /* ansi */
    ;
  }
}

let options = {
  enabled,
  supportLevel
};

function kolorist(start, end, level = 1
/* ansi */
) {
  const open = `\x1b[${start}m`;
  const close = `\x1b[${end}m`;
  const regex = new RegExp(`\\x1b\\[${end}m`, 'g');
  return str => {
    return options.enabled && options.supportLevel >= level ? open + ('' + str).replace(regex, open) + close : '' + str;
  };
}

function stripColors(str) {
  return ('' + str).replace(/\x1b\[[0-9;]+m/g, '').replace(/\x1b\]8;;.*?\x07(.*?)\x1b\]8;;\x07/g, (_, group) => group);
} // modifiers
const bold = kolorist(1, 22);
const dim = kolorist(2, 22);
const red = kolorist(31, 39);
const blue = kolorist(34, 39);
const white = kolorist(97, 39);
const bgRed = kolorist(41, 49);

/**
 * Converts ansi codes into CSS logs that are understood by browsers
 * https://developer.mozilla.org/en-US/docs/Web/API/console#Styling_console_output
 */

const ansiColorsLog = (...input) => {
  // CSS properties, so that nested styles works (in browsers %c will reset styles from any previous %c)
  const styleStack = {};
  const stylesAndSubstitutions = [];
  let str = '';

  for (const segment of input) {
    if (typeof segment !== 'string') {
      stylesAndSubstitutions.push(segment);
      str += '%o'; // https://developer.mozilla.org/en-US/docs/Web/API/console#Using_string_substitutions

      continue;
    }

    str += segment.replace(ansiRegex$1(), escapeCode => {
      // \u001b is unicode for <ESC>
      // eslint-disable-next-line no-control-regex
      const parsedEscapeCode = /\u001B\[(\d*)m/.exec(escapeCode);
      if (!parsedEscapeCode) return ''; // Unrecognized escape code, remove it

      const escapeCodeNum = Number(parsedEscapeCode[1]); // Capture group

      const cssObj = ansiCodes[escapeCodeNum];
      if (!cssObj) return '';
      Object.assign(styleStack, cssObj);
      const cssStr = Object.entries(styleStack).map(([key, val]) => val === 'inherit' ? '' : `${key}: ${val};`).join('');
      stylesAndSubstitutions.push(cssStr);
      return '%c';
    });
  }

  return [str, ...stylesAndSubstitutions];
};
const colors = {
  gray: '#8e908c',
  red: '#c82829',
  green: '#718c00',
  yellow: '#eab700',
  blue: '#4271ae',
  magenta: '#8959a8',
  cyan: '#3e999f',
  orange: '#f5871f',
  white: 'white',
  black: 'black'
};

const fg = color => ({
  color
});

const bg = color => ({
  'background-color': color
});

const ansiCodes = {
  // Styles
  1: {
    'font-weight': 'bold'
  },
  22: {
    'font-weight': 'inherit'
  },
  // Foreground colors
  39: fg('inherit'),
  // Reset foreground
  30: fg(colors.black),
  31: fg(colors.red),
  32: fg(colors.green),
  33: fg(colors.yellow),
  34: fg(colors.blue),
  35: fg(colors.magenta),
  36: fg(colors.cyan),
  37: fg(colors.white),
  97: fg(colors.white),
  // This is the whiteBright color
  90: fg(colors.gray),
  // Background colors
  49: bg('inherit'),
  // Reset background
  40: bg(colors.black),
  41: bg(colors.red),
  42: bg(colors.green),
  43: bg(colors.yellow),
  44: bg(colors.blue),
  45: bg(colors.magenta),
  46: bg(colors.cyan),
  47: bg(colors.white)
};

/**
 * Wraps a JSHandle that points to a function in a browser,
 * with a node function that calls the corresponding browser function,
 * passing along the parameters,
 * and returning the return value wrapped in `Promise<JSHandle<T>>`
 */
const makeCallableJSHandle = browserFunction => (...args) => browserFunction.evaluateHandle((browserFn, ...args) => browserFn(...args), ...args);
const jsHandleToArray = async arrayHandle => {
  const properties = await arrayHandle.getProperties();
  const arr = Array.from({
    length: properties.size
  });

  for (const [i, valHandle] of properties.entries()) {
    // Change primitives to live values rather than JSHandles
    const val = await valHandle.jsonValue().catch(() => valHandle);
    arr[Number(i)] = typeof val === 'object' ? valHandle : val;
  }

  return arr;
};
const isPromise = input => Promise.resolve(input) === input; // https://stackoverflow.com/questions/27746304/how-do-i-tell-if-an-object-is-a-promise/38339199#38339199

const isElementHandle = input => {
  if (typeof input !== 'object' || !input) return false;
  return (input.asElement == null ? void 0 : input.asElement()) === input;
};
const isJSHandle$1 = input => {
  if (typeof input !== 'object' || !input) return false;
  return 'asElement' in input;
};
const assertElementHandle = (input, fn, messageStart = `element parameter must be an ElementHandle\n\n`) => {
  const type = input === null ? 'null' : isPromise(input) ? 'Promise' : typeof input;

  if (type === 'Promise') {
    throw removeFuncFromStackTrace(new Error(`${messageStart}Received Promise. Did you forget await?`), fn);
  }

  if (!isElementHandle(input)) {
    // If it is a JSHandle, that points to something _other_ than an element
    if (isJSHandle$1(input)) {
      throw removeFuncFromStackTrace(new Error(`${messageStart}Received a JSHandle that did not point to an element`), fn);
    }

    throw removeFuncFromStackTrace(new Error(`${messageStart}Received ${type}`), fn);
  }
};
const changeErrorMessage = (error, changeMessage) => {
  const newMessage = changeMessage(error.message);
  if (error.stack) error.stack = error.stack.replace(error.message, newMessage);
  error.message = newMessage;
  return error;
};
/**
 * Manipulate the stack trace and remove fn from it
 * That way jest will show a code frame from the user's code, not ours
 * https://kentcdodds.com/blog/improve-test-error-messages-of-your-abstractions
 */

const removeFuncFromStackTrace = (error, fn) => {
  Error.captureStackTrace(error, fn);
  return error;
};
const printStackLine = (path, line, column, fnName) => {
  const location = fnName ? `${fnName} (${path}:${line}:${column})` : `${path}:${line}:${column}`;
  return `    at ${location}`;
};
const printColorsInErrorMessages = // The PLEASANTEST_TESTING_ITSELF environment variable (set in jest.setup.ts)
// allows us to detect whether PT is running in the context of its own tests
// We use it to disable syntax-highlighting in printed HTML in error messages
// so that the snapshots are more readable.
// When PT is used outside of its own tests, the environment variable will not be set,
// so the error messages will have syntax-highlighted HTML
options.enabled && process.env.PLEASANTEST_TESTING_ITSELF !== 'true';

/**
 * Manages asynchronous calls within withBrowser.
 * If any async calls are "left over" when withBrowser finishes executing,
 * then that means the user forgot to await the async calls,
 * so we should throw an error that indicates that.
 */
/**
 * Set of all active async hook trackers
 * We need to store this module-level so that jest-dom matchers and getAccessibilityTree
 * can know which withBrowser they "belong" to
 * If there are multiple active at a time, the jest-dom matchers won't include the forgot-await behavior.
 */

const activeAsyncHookTrackers = new Set();
const createAsyncHookTracker = () => {
  const hooks = new Set();
  let isClosed = false;

  const addHook = async (func, captureFunction) => {
    const forgotAwaitError = new Error('Cannot interact with browser after test finishes. Did you forget to await?');
    removeFuncFromStackTrace(forgotAwaitError, captureFunction);
    hooks.add(forgotAwaitError);

    try {
      return await func();
    } catch (error) {
      // If we throw an error here and it _is_ closed,
      // there will be an unhandled rejection, ending the process, without a code frame.
      //
      // If it is closed, it is better to not throw an error because when close() was called,
      // we would have already noticed that there was an async hook and thrown an error there.
      // So, we only throw an error if it is open
      if (!isClosed) throw error; // This line has no runtime effect, it is just there to make TS OK
      // If someone forgot await and is using promise-wrapped values directly,
      // TS is already giving them useful error messages.
      // The `as never` here tells TS that we can ignore that sometimes this will return undefined

      return undefined;
    } finally {
      if (!isClosed) hooks.delete(forgotAwaitError);
    }
  };

  const close = () => {
    if (isClosed) return;
    isClosed = true;
    activeAsyncHookTrackers.delete(asyncHookTracker);
    if (hooks.size > 0) return hooks[Symbol.iterator]().next().value;
  };

  const asyncHookTracker = {
    addHook,
    close
  };
  activeAsyncHookTrackers.add(asyncHookTracker);
  return asyncHookTracker;
};

var startDisownedBrowserPath = require('path').join(__dirname,'start-disowned-browser.cjs');

const installFolder$1 = path__namespace.dirname(path__namespace.dirname(path__namespace.dirname(node_url.fileURLToPath(`file://${__filename}`)))); // Something like <something>/node_modules/pleasantest/.browser-cache.json

const cachePath = path__namespace.join(installFolder$1, '.browser-cache.json');

const readCache = async () => {
  try {
    const cache = await node_fs.promises.readFile(cachePath, 'utf8').catch(() => '');
    const parsed = JSON.parse(cache);
    if (typeof parsed === 'object') return parsed;
  } catch {}

  return {};
};

const updateCacheFile = async (browser, headless, value, previousValue) => {
  await node_fs.promises.mkdir(path__namespace.dirname(cachePath), {
    recursive: true
  });
  const oldCache = await readCache();
  const headlessStr = headless ? 'headless' : 'headed';
  const browserObj = oldCache[browser] || (oldCache[browser] = {});

  if (previousValue !== undefined && previousValue !== browserObj[headlessStr]) {
    return browserObj[headlessStr];
  }

  browserObj[headlessStr] = value;
  await node_fs.promises.writeFile(cachePath, JSON.stringify(oldCache, null, 2));
};

const connectToCachedBrowser = async (browser, headless, timeLimit = 5000) => {
  var _cache$browser;

  const cache = await readCache();
  const cachedWSEndpoint = (_cache$browser = cache[browser]) == null ? void 0 : _cache$browser[headless ? 'headless' : 'headed']; // In case another process is currently starting a browser, wait for that process
  // rather than starting a whole new one

  if (cachedWSEndpoint === 'starting' && timeLimit > 0) {
    return new Promise(resolve => {
      // Every 50ms check again (this is recursive)
      setTimeout(() => connectToCachedBrowser(browser, headless, timeLimit - 50).then(resolve), 50);
    });
  }

  if (cachedWSEndpoint) {
    return puppeteer.connect({
      browserWSEndpoint: cachedWSEndpoint
    }).catch(() => ({
      connected: false,
      previousValue: cachedWSEndpoint
    }));
  }

  return {
    connected: false,
    previousValue: cachedWSEndpoint
  };
};

const isBrowser = input => // @ts-expect-error checking for properties on unknown object
input && typeof input === 'object' && input.version;

const connectToBrowser = async (browser, headless) => {
  // I acknowledge that this code is gross and should be refactored
  // Constraints:
  // - If there is no browser in the cache, multiple concurrent processes should only start 1 new browser
  // - If there is a killed browser in the cache, multiple concurrent processes should only start 1 new browser
  // - If there "starting" in the cache but nothing is really starting, multiple concurrent processes should only start 1 new browser
  // TODO: Idea: use a state machine!!!
  const cachedBrowser = await connectToCachedBrowser(browser, headless);

  if (isBrowser(cachedBrowser)) {
    return cachedBrowser;
  }

  let valueWrittenInMeantime = await updateCacheFile(browser, headless, 'starting', cachedBrowser.previousValue);

  if (valueWrittenInMeantime) {
    const connectedBrowser = await connectToCachedBrowser(browser, headless);
    if (!isBrowser(connectedBrowser)) throw new Error('unable to connect to browser');
    return connectedBrowser;
  }

  const subprocess = childProcess__namespace.fork(startDisownedBrowserPath, {
    detached: true,
    stdio: 'ignore'
  });
  const wsEndpoint = await new Promise((resolve, reject) => {
    subprocess.send({
      browser,
      headless
    });
    subprocess.on('message', msg => {
      if (msg.error) {
        reject(new Error(`Failed to start browser: ${msg.error}`));
        return;
      }

      if (!msg.browserWSEndpoint) return;
      resolve(msg.browserWSEndpoint);
    });
  }).catch(async error => {
    subprocess.kill();
    valueWrittenInMeantime = await updateCacheFile(browser, headless, '', 'starting');
    throw error;
  });
  valueWrittenInMeantime = await updateCacheFile(browser, headless, wsEndpoint, 'starting');

  if (valueWrittenInMeantime) {
    // Another browser was started while this browser was starting
    // so we are going to kill the current browser and connect to the other one instead
    subprocess.kill();
    return puppeteer.connect({
      browserWSEndpoint: valueWrittenInMeantime
    });
  } // Disconnect from the spawned process so it can keep running in the background


  subprocess.unref();
  subprocess.disconnect();
  return puppeteer.connect({
    browserWSEndpoint: wsEndpoint
  });
};

// This makes it so that errors thrown by the module server get re-thrown inside of runJS/loadJS.
// This is necessary because the network is between the runJS call and the module server, so errors do not propagate
// By tracking which runJS/loadJS initiated the request for each file, the module server can know which runJS/loadJS needs to reject.
// When runJS finishes, it will check to see if the buildStatuses map has any errors corresponding to its buildId.
// If there are any errors, it throws the first one.
const buildStatuses = new Map();
let buildIds = 0;
/**
 * Add an error for a specific buildId.
 * If the build is already finished (resolved),
 * triggers an uncaught rejection, with the hopes that it will fail the test.
 */

const rejectBuild = (buildId, error) => {
  const statusArray = buildStatuses.get(buildId);
  if (statusArray) statusArray.push(error); // Uncaught promise rejection!
  // Hope that Jest will catch it and fail the test, otherwise it is just logged by Node
  else Promise.reject(error);
};
const createBuildStatusTracker = () => {
  const buildId = ++buildIds;
  buildStatuses.set(buildId, []);
  return {
    buildId,

    complete() {
      const status = buildStatuses.get(buildId); // This should never happen

      if (!status) throw new Error('Build already completed');
      buildStatuses.delete(buildId);
      if (status.length > 0) return status;
    }

  };
};

function every (arr, cb) {
  var i = 0,
      len = arr.length;

  for (; i < len; i++) {
    if (!cb(arr[i], i, arr)) {
      return false;
    }
  }

  return true;
}

const SEP = '/'; // Types ~> static, param, any, optional

const STYPE = 0,
      PTYPE = 1,
      ATYPE = 2,
      OTYPE = 3; // Char Codes ~> / : *

const SLASH = 47,
      COLON = 58,
      ASTER = 42,
      QMARK = 63;

function strip(str) {
  if (str === SEP) return str;
  str.charCodeAt(0) === SLASH && (str = str.substring(1));
  var len = str.length - 1;
  return str.charCodeAt(len) === SLASH ? str.substring(0, len) : str;
}

function split(str) {
  return (str = strip(str)) === SEP ? [SEP] : str.split(SEP);
}

function isMatch(arr, obj, idx) {
  idx = arr[idx];
  return obj.val === idx && obj.type === STYPE || (idx === SEP ? obj.type > PTYPE : obj.type !== STYPE && (idx || '').endsWith(obj.end));
}

function match(str, all) {
  var i = 0,
      tmp,
      segs = split(str),
      len = segs.length,
      l;
  var fn = isMatch.bind(isMatch, segs);

  for (; i < all.length; i++) {
    tmp = all[i];

    if ((l = tmp.length) === len || l < len && tmp[l - 1].type === ATYPE || l > len && tmp[l - 1].type === OTYPE) {
      if (every(tmp, fn)) return tmp;
    }
  }

  return [];
}
function parse(str) {
  if (str === SEP) {
    return [{
      old: str,
      type: STYPE,
      val: str,
      end: ''
    }];
  }

  var c,
      x,
      t,
      sfx,
      nxt = strip(str),
      i = -1,
      j = 0,
      len = nxt.length,
      out = [];

  while (++i < len) {
    c = nxt.charCodeAt(i);

    if (c === COLON) {
      j = i + 1; // begining of param

      t = PTYPE; // set type

      x = 0; // reset mark

      sfx = '';

      while (i < len && nxt.charCodeAt(i) !== SLASH) {
        c = nxt.charCodeAt(i);

        if (c === QMARK) {
          x = i;
          t = OTYPE;
        } else if (c === 46 && sfx.length === 0) {
          sfx = nxt.substring(x = i);
        }

        i++; // move on
      }

      out.push({
        old: str,
        type: t,
        val: nxt.substring(j, x || i),
        end: sfx
      }); // shorten string & update pointers

      nxt = nxt.substring(i);
      len -= i;
      i = 0;
      continue; // loop
    } else if (c === ASTER) {
      out.push({
        old: str,
        type: ATYPE,
        val: nxt.substring(i),
        end: ''
      });
      continue; // loop
    } else {
      j = i;

      while (i < len && nxt.charCodeAt(i) !== SLASH) {
        ++i; // skip to next slash
      }

      out.push({
        old: str,
        type: STYPE,
        val: nxt.substring(j, i),
        end: ''
      }); // shorten string & update pointers

      nxt = nxt.substring(i);
      len -= i;
      i = j = 0;
    }
  }

  return out;
}
function exec(str, arr) {
  var i = 0,
      x,
      y,
      segs = split(str),
      out = {};

  for (; i < arr.length; i++) {
    x = segs[i];
    y = arr[i];
    if (x === SEP) continue;

    if (x !== void 0 && y.type | 2 === OTYPE) {
      out[y.val] = x.replace(y.end, '');
    }
  }

  return out;
}

class Trouter {
  constructor(opts) {
    this.opts = opts || {};
    this.routes = {};
    this.handlers = {};
    this.all = this.add.bind(this, '*');
    this.get = this.add.bind(this, 'GET');
    this.head = this.add.bind(this, 'HEAD');
    this.patch = this.add.bind(this, 'PATCH');
    this.options = this.add.bind(this, 'OPTIONS');
    this.connect = this.add.bind(this, 'CONNECT');
    this.delete = this.add.bind(this, 'DELETE');
    this.trace = this.add.bind(this, 'TRACE');
    this.post = this.add.bind(this, 'POST');
    this.put = this.add.bind(this, 'PUT');
  }

  add(method, pattern, ...fns) {
    // Save decoded pattern info
    if (this.routes[method] === void 0) this.routes[method] = [];
    this.routes[method].push(parse(pattern)); // Save route handler(s)

    if (this.handlers[method] === void 0) this.handlers[method] = {};
    this.handlers[method][pattern] = fns; // Allow chainable

    return this;
  }

  find(method, url) {
    let arr = match(url, this.routes[method] || []);

    if (arr.length === 0) {
      arr = match(url, this.routes[method = '*'] || []);
      if (!arr.length) return false;
    }

    return {
      params: exec(url, arr),
      handlers: this.handlers[method][arr[0].old]
    };
  }

}

let _exports$1 = function (req) {
  let url = req.url;
  if (url === void 0) return url;
  let obj = req._parsedUrl;
  if (obj && obj._raw === url) return obj;
  obj = {};
  obj.query = obj.search = null;
  obj.href = obj.path = obj.pathname = url;
  let idx = url.indexOf('?', 1);

  if (idx !== -1) {
    obj.search = url.substring(idx);
    obj.query = obj.search.substring(1);
    obj.pathname = url.substring(0, idx);
  }

  obj._raw = url;
  return req._parsedUrl = obj;
};

function lead(x) {
  return x.charCodeAt(0) === 47 ? x : '/' + x;
}

function value(x) {
  let y = x.indexOf('/', 1);
  return y > 1 ? x.substring(0, y) : x;
}

function mutate(str, req) {
  req.url = req.url.substring(str.length) || '/';
  req.path = req.path.substring(str.length) || '/';
}

function onError(err, req, res, next) {
  let code = res.statusCode = err.code || err.status || 500;
  res.end(err.length && err || err.message || http__namespace.STATUS_CODES[code]);
}

class Polka extends Trouter {
  constructor(opts = {}) {
    super(opts);
    this.apps = {};
    this.wares = [];
    this.bwares = {};
    this.parse = _exports$1;
    this.server = opts.server;
    this.handler = this.handler.bind(this);
    this.onError = opts.onError || onError; // catch-all handler

    this.onNoMatch = opts.onNoMatch || this.onError.bind(null, {
      code: 404
    });
  }

  add(method, pattern, ...fns) {
    let base = lead(value(pattern));
    if (this.apps[base] !== void 0) throw new Error(`Cannot mount ".${method.toLowerCase()}('${lead(pattern)}')" because a Polka application at ".use('${base}')" already exists! You should move this handler into your Polka application instead.`);
    return super.add(method, pattern, ...fns);
  }

  use(base, ...fns) {
    if (typeof base === 'function') {
      this.wares = this.wares.concat(base, fns);
    } else if (base === '/') {
      this.wares = this.wares.concat(fns);
    } else {
      base = lead(base);
      fns.forEach(fn => {
        if (fn instanceof Polka) {
          this.apps[base] = fn;
        } else {
          let arr = this.bwares[base] || [];
          arr.length > 0 || arr.push((r, _, nxt) => (mutate(base, r), nxt()));
          this.bwares[base] = arr.concat(fn);
        }
      });
    }

    return this; // chainable
  }

  listen() {
    (this.server = this.server || http__namespace.createServer()).on('request', this.handler);
    this.server.listen.apply(this.server, arguments);
    return this;
  }

  handler(req, res, info) {
    info = info || this.parse(req);
    let fns = [],
        arr = this.wares,
        obj = this.find(req.method, info.pathname);
    req.originalUrl = req.originalUrl || req.url;
    let base = value(req.path = info.pathname);

    if (this.bwares[base] !== void 0) {
      arr = arr.concat(this.bwares[base]);
    }

    if (obj) {
      fns = obj.handlers;
      req.params = obj.params;
    } else if (this.apps[base] !== void 0) {
      mutate(base, req);
      info.pathname = req.path; //=> updates

      fns.push(this.apps[base].handler.bind(null, req, res, info));
    } else if (fns.length === 0) {
      fns.push(this.onNoMatch);
    } // Grab addl values from `info`


    req.search = info.search;
    req.query = querystring.parse(info.query); // Exit if only a single function

    let i = 0,
        len = arr.length,
        num = fns.length;
    if (len === i && num === 1) return fns[0](req, res); // Otherwise loop thru all middlware

    let next = err => err ? this.onError(err, req, res, next) : loop();

    let loop = _ => res.finished || i < len && arr[i++](req, res, next);

    arr = arr.concat(fns);
    len += num;
    loop(); // init
  }

}

let _exports = opts => new Polka(opts);

const createServer = ({
  middleware
}) => new Promise(resolve => {
  const server = _exports({
    onError(err, req, res) {
      const fullPath = req.originalUrl.replace(/\?.+$/, ''); // ignore missing favicon requests

      if (fullPath === '/favicon.ico') {
        res.writeHead(200, {
          'content-type': 'image/x-icon',
          'content-length': '0'
        });
        return res.end('');
      } // @ts-expect-error TS doesn't know about err.code


      const code = typeof err.code === 'number' ? err.code : 500;
      res.statusCode = code;
      res.writeHead(code, {
        'content-type': 'text/plain'
      });
      if (code === 404) return res.end('not found');
      res.end(err.stack); // Create a new console instance instead of using the global one
      // Because the global one is overridden by Jest, and it adds a misleading second stack trace and code frame below it

      const console = new node_console.Console(process.stdout, process.stderr);
      console.log(err.stack || err.message || err);
    }

  });
  if (middleware.length > 0) server.use(...middleware);
  const sockets = new Set();
  server.listen( // 0 means the OS will choose a random free port
  0, undefined, () => {
    const serv = server.server;
    serv.on('connection', socket => {
      sockets.add(socket);
      socket.on('close', () => sockets.delete(socket));
    });
    resolve({
      port: server.server.address().port,
      server,
      close: async () => {
        for (const socket of sockets) socket.destroy();

        await new Promise((resolve, reject) => {
          serv.close(err => err ? reject(err) : resolve());
        });
      }
    });
  });
});

let currentDir;

try {
  currentDir = path__default.dirname(node_url.fileURLToPath(`file://${__filename}`));
} catch {
  currentDir = __dirname;
}

const cache = new Map();

const clientRuntimeMiddleware = (root = path__default.resolve(currentDir, '..')) => async (req, res, next) => {
  if (!req.path.startsWith('/@pleasantest')) {
    next();
    return;
  }

  res.setHeader('Content-Type', 'application/javascript;charset=utf-8');
  res.setHeader('Access-Control-Allow-Origin', '*');
  const cached = cache.get(req.path);

  if (cached) {
    res.end(cached);
    return;
  }

  const filePath = path__default.join(root, req.path === '/@pleasantest/jest-dom' ? 'jest-dom.js' : req.path === '/@pleasantest/user-util' ? 'user-util.js' : req.path === '/@pleasantest/accessibility' ? 'accessibility.js' : 'pptr-testing-library-client.js');
  const text = await node_fs.promises.readFile(filePath, 'utf8');
  res.end(text);
  cache.set(req.path, text);
};

let cachedServerPromise;
/**
 * The Client Runtime server serves up the bundles for testing library, jest-dom, and user-event
 * The instance is reused (there should only ever be once instance per node process)
 */

const createClientRuntimeServer = async rootDir => {
  if (cachedServerPromise) return cachedServerPromise;
  return cachedServerPromise = createServer({
    middleware: [clientRuntimeMiddleware(rootDir)]
  });
};
const cleanupClientRuntimeServer = async () => {
  const cachedServer = await cachedServerPromise;
  await (cachedServer == null ? void 0 : cachedServer.close());
};

const npmPrefix = '@npm/';
const isRelativeOrAbsoluteImport = id => id === '.' || id === '..' || id.startsWith('./') || id.startsWith('../') || id.startsWith('/');
const isBareImport = id => !(isRelativeOrAbsoluteImport(id) || id.startsWith('\0') || id.startsWith(npmPrefix));
const cssExts = /\.(?:css|styl|stylus|s[ac]ss|less)$/;
const jsExts = /\.(?:[jt]sx?|[cm]js)$/;

// Copied from https://github.com/preactjs/wmr/blob/18b8f00923ddf578f2f747c35bd24a8039eee3ba/packages/wmr/src/lib/transform-css-imports.js

/*
  https://github.com/preactjs/wmr/blob/main/LICENSE
  MIT License
  Copyright (c) 2020 The Preact Authors
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
  */

/*
  Differences from original:
  - Types
  - ESLint fixes
  */

/**
 * @param code Module code
 * @param id Source module specifier
 */
const transformCssImports = async (code, id, {
  resolveId
}) => {
  const CSS_IMPORTS = /@import\s+["'](.*?)["'];|url\(["']?(.*?)["']?\)/g;
  let out = code;
  let offset = 0;
  let match;

  while (match = CSS_IMPORTS.exec(code)) {
    const spec = match[1] || match[2];
    const start = match.index + match[0].indexOf(spec) + offset;
    const end = start + spec.length;
    const resolved = await resolveId(spec, id);

    if (typeof resolved === 'string') {
      out = out.slice(0, start) + resolved + out.slice(end);
      offset += resolved.length - spec.length;
    }
  }

  return out;
};

const cssPlugin = ({
  returnCSS = false,
  root
}) => {
  const transformedCSS = new Map();
  const plugin = postcssPlugin__default({
    inject: cssVariable => `
        const style = document.createElement('style')
        style.type = 'text/css'
        const promise = new Promise(r => style.addEventListener('load', r))
        style.appendChild(document.createTextNode(${cssVariable}))
        document.head.append(style)
        await promise
      `,
    // They are executed right to left. We want our custom loader to run last
    use: ['rewriteImports', 'sass', 'stylus', 'less'],
    loaders: [{
      // Rewrites emitted url(...) and @imports to be relative to the project root.
      // Otherwise, relative paths don't work for injected stylesheets
      name: 'rewriteImports',
      test: cssExts,

      async process({
        code,
        map
      }) {
        code = await transformCssImports(code, this.id, {
          resolveId: (spec, id) => {
            if (!spec.startsWith('./')) return spec.split(path.sep).join(path.posix.sep);
            const absolutePath = path.resolve(id, '..', spec);
            return `./${path.relative(root, absolutePath).split(path.sep).join(path.posix.sep)}`;
          }
        });

        if (returnCSS) {
          transformedCSS.set(this.id, code);
        }

        return {
          code,
          map
        };
      }

    }]
  }); // Adds .meta.css to returned object (for use in CSS middleware)

  if (returnCSS) {
    const originalTranform = 'handler' in plugin.transform ? plugin.transform.handler : plugin.transform;

    plugin.transform = async function (code, id) {
      let result = await originalTranform.call(this, code, id);
      if (typeof result === 'string') result = {
        code: result,
        map: ''
      };else if (!result) return result;
      if (!result.meta) result.meta = {};
      result.meta.css = transformedCSS.get(id);
      return result;
    };
  }

  return plugin;
};

/**
 * This middleware handles _only_ css files that were _not imported from JS_
 * CSS files that were imported from JS have the ?import param and were handled by the JS middleware (transformed to JS)
 * This middleware does not transform CSS to JS that can be imported, it just returns CSS
 * Use cases: CSS included via <link> tags, CSS included via @import
 * TODO: consider using this for loadCSS
 */
const cssMiddleware = ({
  root
}) => {
  const cssPlug = cssPlugin({
    root,
    returnCSS: true
  });
  return async (req, res, next) => {
    try {
      if (!cssExts.test(req.path)) {
        next();
        return;
      } // Normalized path starting with slash


      const path$1 = path.posix.normalize(req.path); // Remove leading slash, and convert slashes to os-specific slashes

      const osPath = path$1.slice(1).split(path.posix.sep).join(path.sep); // Absolute file path

      const file = path.resolve(root, osPath); // Rollup-style CWD-relative Unix-normalized path "id":

      const id = `./${path.relative(root, file).replace(/^\.\//, '').replace(/^\0/, '').split(path.sep).join(path.posix.sep)}`;
      res.setHeader('Content-Type', 'text/css;charset=utf-8');
      let code = await node_fs.promises.readFile(file, 'utf8');

      if (cssPlug.transform) {
        var _result$meta;

        const ctx = {
          warn(...args) {
            console.log(`[${cssPlug.name}]`, ...args);
          },

          error(error) {
            if (typeof error === 'string') throw new Error(error);
            throw error;
          }

        }; // We need to call the transform hook, but get the CSS out of it before it converts it to JS

        const cssTransform = 'handler' in cssPlug.transform ? cssPlug.transform.handler : cssPlug.transform;
        const result = await cssTransform.call(ctx, code, id);

        if (typeof result !== 'object' || result === null || ((_result$meta = result.meta) == null ? void 0 : _result$meta.css) === undefined) {
          next();
          return;
        }

        code = result.meta.css;
      }

      if (!code) {
        next();
        return;
      }

      res.writeHead(200, {
        'Content-Length': Buffer.byteLength(code, 'utf8')
      });
      res.end(code);
    } catch (error) {
      next(error);
    }
  };
};

const defaultHTML = bodyContent => `
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <link rel="icon" href="data:;base64,=" />
    <title>pleasantest</title>
  </head>
  <body>${bodyContent}</body>
</html>
`;
const indexHTMLMiddleware = (req, res, next) => {
  if (req.url !== '/') {
    next();
    return;
  }

  res.statusCode = 200;
  res.setHeader('Content-Type', 'text/html');
  res.write(defaultHTML(`<h1 style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%)">
        Your test will run here
      </h1>`));
  res.end();
};

const comma = ','.charCodeAt(0);
const semicolon = ';'.charCodeAt(0);
const chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
const intToChar = new Uint8Array(64); // 64 possible chars.

const charToInt = new Uint8Array(128); // z is 122 in ASCII

for (let i = 0; i < chars.length; i++) {
  const c = chars.charCodeAt(i);
  intToChar[i] = c;
  charToInt[c] = i;
} // Provide a fallback for older environments.


const td = typeof TextDecoder !== 'undefined' ? /* #__PURE__ */new TextDecoder() : typeof Buffer !== 'undefined' ? {
  decode(buf) {
    const out = Buffer.from(buf.buffer, buf.byteOffset, buf.byteLength);
    return out.toString();
  }

} : {
  decode(buf) {
    let out = '';

    for (let i = 0; i < buf.length; i++) {
      out += String.fromCharCode(buf[i]);
    }

    return out;
  }

};

function decode(mappings) {
  const state = new Int32Array(5);
  const decoded = [];
  let index = 0;

  do {
    const semi = indexOf(mappings, index);
    const line = [];
    let sorted = true;
    let lastCol = 0;
    state[0] = 0;

    for (let i = index; i < semi; i++) {
      let seg;
      i = decodeInteger(mappings, i, state, 0); // genColumn

      const col = state[0];
      if (col < lastCol) sorted = false;
      lastCol = col;

      if (hasMoreVlq(mappings, i, semi)) {
        i = decodeInteger(mappings, i, state, 1); // sourcesIndex

        i = decodeInteger(mappings, i, state, 2); // sourceLine

        i = decodeInteger(mappings, i, state, 3); // sourceColumn

        if (hasMoreVlq(mappings, i, semi)) {
          i = decodeInteger(mappings, i, state, 4); // namesIndex

          seg = [col, state[1], state[2], state[3], state[4]];
        } else {
          seg = [col, state[1], state[2], state[3]];
        }
      } else {
        seg = [col];
      }

      line.push(seg);
    }

    if (!sorted) sort(line);
    decoded.push(line);
    index = semi + 1;
  } while (index <= mappings.length);

  return decoded;
}

function indexOf(mappings, index) {
  const idx = mappings.indexOf(';', index);
  return idx === -1 ? mappings.length : idx;
}

function decodeInteger(mappings, pos, state, j) {
  let value = 0;
  let shift = 0;
  let integer = 0;

  do {
    const c = mappings.charCodeAt(pos++);
    integer = charToInt[c];
    value |= (integer & 31) << shift;
    shift += 5;
  } while (integer & 32);

  const shouldNegate = value & 1;
  value >>>= 1;

  if (shouldNegate) {
    value = -0x80000000 | -value;
  }

  state[j] += value;
  return pos;
}

function hasMoreVlq(mappings, i, length) {
  if (i >= length) return false;
  return mappings.charCodeAt(i) !== comma;
}

function sort(line) {
  line.sort(sortComparator$1);
}

function sortComparator$1(a, b) {
  return a[0] - b[0];
}

function encode(decoded) {
  const state = new Int32Array(5);
  const bufLength = 1024 * 16;
  const subLength = bufLength - 36;
  const buf = new Uint8Array(bufLength);
  const sub = buf.subarray(0, subLength);
  let pos = 0;
  let out = '';

  for (let i = 0; i < decoded.length; i++) {
    const line = decoded[i];

    if (i > 0) {
      if (pos === bufLength) {
        out += td.decode(buf);
        pos = 0;
      }

      buf[pos++] = semicolon;
    }

    if (line.length === 0) continue;
    state[0] = 0;

    for (let j = 0; j < line.length; j++) {
      const segment = line[j]; // We can push up to 5 ints, each int can take at most 7 chars, and we
      // may push a comma.

      if (pos > subLength) {
        out += td.decode(sub);
        buf.copyWithin(0, subLength, pos);
        pos -= subLength;
      }

      if (j > 0) buf[pos++] = comma;
      pos = encodeInteger(buf, pos, state, segment, 0); // genColumn

      if (segment.length === 1) continue;
      pos = encodeInteger(buf, pos, state, segment, 1); // sourcesIndex

      pos = encodeInteger(buf, pos, state, segment, 2); // sourceLine

      pos = encodeInteger(buf, pos, state, segment, 3); // sourceColumn

      if (segment.length === 4) continue;
      pos = encodeInteger(buf, pos, state, segment, 4); // namesIndex
    }
  }

  return out + td.decode(buf.subarray(0, pos));
}

function encodeInteger(buf, pos, state, segment, j) {
  const next = segment[j];
  let num = next - state[j];
  state[j] = next;
  num = num < 0 ? -num << 1 | 1 : num << 1;

  do {
    let clamped = num & 0b011111;
    num >>>= 5;
    if (num > 0) clamped |= 0b100000;
    buf[pos++] = intToChar[clamped];
  } while (num > 0);

  return pos;
}

class BitSet {
  constructor(arg) {
    this.bits = arg instanceof BitSet ? arg.bits.slice() : [];
  }

  add(n) {
    this.bits[n >> 5] |= 1 << (n & 31);
  }

  has(n) {
    return !!(this.bits[n >> 5] & 1 << (n & 31));
  }

}

class Chunk {
  constructor(start, end, content) {
    this.start = start;
    this.end = end;
    this.original = content;
    this.intro = '';
    this.outro = '';
    this.content = content;
    this.storeName = false;
    this.edited = false;
    {
      this.previous = null;
      this.next = null;
    }
  }

  appendLeft(content) {
    this.outro += content;
  }

  appendRight(content) {
    this.intro = this.intro + content;
  }

  clone() {
    const chunk = new Chunk(this.start, this.end, this.original);
    chunk.intro = this.intro;
    chunk.outro = this.outro;
    chunk.content = this.content;
    chunk.storeName = this.storeName;
    chunk.edited = this.edited;
    return chunk;
  }

  contains(index) {
    return this.start < index && index < this.end;
  }

  eachNext(fn) {
    let chunk = this;

    while (chunk) {
      fn(chunk);
      chunk = chunk.next;
    }
  }

  eachPrevious(fn) {
    let chunk = this;

    while (chunk) {
      fn(chunk);
      chunk = chunk.previous;
    }
  }

  edit(content, storeName, contentOnly) {
    this.content = content;

    if (!contentOnly) {
      this.intro = '';
      this.outro = '';
    }

    this.storeName = storeName;
    this.edited = true;
    return this;
  }

  prependLeft(content) {
    this.outro = content + this.outro;
  }

  prependRight(content) {
    this.intro = content + this.intro;
  }

  split(index) {
    const sliceIndex = index - this.start;
    const originalBefore = this.original.slice(0, sliceIndex);
    const originalAfter = this.original.slice(sliceIndex);
    this.original = originalBefore;
    const newChunk = new Chunk(index, this.end, originalAfter);
    newChunk.outro = this.outro;
    this.outro = '';
    this.end = index;

    if (this.edited) {
      // TODO is this block necessary?...
      newChunk.edit('', false);
      this.content = '';
    } else {
      this.content = originalBefore;
    }

    newChunk.next = this.next;
    if (newChunk.next) newChunk.next.previous = newChunk;
    newChunk.previous = this;
    this.next = newChunk;
    return newChunk;
  }

  toString() {
    return this.intro + this.content + this.outro;
  }

  trimEnd(rx) {
    this.outro = this.outro.replace(rx, '');
    if (this.outro.length) return true;
    const trimmed = this.content.replace(rx, '');

    if (trimmed.length) {
      if (trimmed !== this.content) {
        this.split(this.start + trimmed.length).edit('', undefined, true);
      }

      return true;
    } else {
      this.edit('', undefined, true);
      this.intro = this.intro.replace(rx, '');
      if (this.intro.length) return true;
    }
  }

  trimStart(rx) {
    this.intro = this.intro.replace(rx, '');
    if (this.intro.length) return true;
    const trimmed = this.content.replace(rx, '');

    if (trimmed.length) {
      if (trimmed !== this.content) {
        this.split(this.end - trimmed.length);
        this.edit('', undefined, true);
      }

      return true;
    } else {
      this.edit('', undefined, true);
      this.outro = this.outro.replace(rx, '');
      if (this.outro.length) return true;
    }
  }

}

function getBtoa() {
  if (typeof window !== 'undefined' && typeof window.btoa === 'function') {
    return str => window.btoa(unescape(encodeURIComponent(str)));
  } else if (typeof Buffer === 'function') {
    return str => Buffer.from(str, 'utf-8').toString('base64');
  } else {
    return () => {
      throw new Error('Unsupported environment: `window.btoa` or `Buffer` should be supported.');
    };
  }
}

const btoa = /*#__PURE__*/getBtoa();

class SourceMap$1 {
  constructor(properties) {
    this.version = 3;
    this.file = properties.file;
    this.sources = properties.sources;
    this.sourcesContent = properties.sourcesContent;
    this.names = properties.names;
    this.mappings = encode(properties.mappings);
  }

  toString() {
    return JSON.stringify(this);
  }

  toUrl() {
    return 'data:application/json;charset=utf-8;base64,' + btoa(this.toString());
  }

}

function guessIndent(code) {
  const lines = code.split('\n');
  const tabbed = lines.filter(line => /^\t+/.test(line));
  const spaced = lines.filter(line => /^ {2,}/.test(line));

  if (tabbed.length === 0 && spaced.length === 0) {
    return null;
  } // More lines tabbed than spaced? Assume tabs, and
  // default to tabs in the case of a tie (or nothing
  // to go on)


  if (tabbed.length >= spaced.length) {
    return '\t';
  } // Otherwise, we need to guess the multiple


  const min = spaced.reduce((previous, current) => {
    const numSpaces = /^ +/.exec(current)[0].length;
    return Math.min(numSpaces, previous);
  }, Infinity);
  return new Array(min + 1).join(' ');
}

function getRelativePath(from, to) {
  const fromParts = from.split(/[/\\]/);
  const toParts = to.split(/[/\\]/);
  fromParts.pop(); // get dirname

  while (fromParts[0] === toParts[0]) {
    fromParts.shift();
    toParts.shift();
  }

  if (fromParts.length) {
    let i = fromParts.length;

    while (i--) fromParts[i] = '..';
  }

  return fromParts.concat(toParts).join('/');
}

const toString = Object.prototype.toString;

function isObject(thing) {
  return toString.call(thing) === '[object Object]';
}

function getLocator(source) {
  const originalLines = source.split('\n');
  const lineOffsets = [];

  for (let i = 0, pos = 0; i < originalLines.length; i++) {
    lineOffsets.push(pos);
    pos += originalLines[i].length + 1;
  }

  return function locate(index) {
    let i = 0;
    let j = lineOffsets.length;

    while (i < j) {
      const m = i + j >> 1;

      if (index < lineOffsets[m]) {
        j = m;
      } else {
        i = m + 1;
      }
    }

    const line = i - 1;
    const column = index - lineOffsets[line];
    return {
      line,
      column
    };
  };
}

class Mappings {
  constructor(hires) {
    this.hires = hires;
    this.generatedCodeLine = 0;
    this.generatedCodeColumn = 0;
    this.raw = [];
    this.rawSegments = this.raw[this.generatedCodeLine] = [];
    this.pending = null;
  }

  addEdit(sourceIndex, content, loc, nameIndex) {
    if (content.length) {
      const segment = [this.generatedCodeColumn, sourceIndex, loc.line, loc.column];

      if (nameIndex >= 0) {
        segment.push(nameIndex);
      }

      this.rawSegments.push(segment);
    } else if (this.pending) {
      this.rawSegments.push(this.pending);
    }

    this.advance(content);
    this.pending = null;
  }

  addUneditedChunk(sourceIndex, chunk, original, loc, sourcemapLocations) {
    let originalCharIndex = chunk.start;
    let first = true;

    while (originalCharIndex < chunk.end) {
      if (this.hires || first || sourcemapLocations.has(originalCharIndex)) {
        this.rawSegments.push([this.generatedCodeColumn, sourceIndex, loc.line, loc.column]);
      }

      if (original[originalCharIndex] === '\n') {
        loc.line += 1;
        loc.column = 0;
        this.generatedCodeLine += 1;
        this.raw[this.generatedCodeLine] = this.rawSegments = [];
        this.generatedCodeColumn = 0;
        first = true;
      } else {
        loc.column += 1;
        this.generatedCodeColumn += 1;
        first = false;
      }

      originalCharIndex += 1;
    }

    this.pending = null;
  }

  advance(str) {
    if (!str) return;
    const lines = str.split('\n');

    if (lines.length > 1) {
      for (let i = 0; i < lines.length - 1; i++) {
        this.generatedCodeLine++;
        this.raw[this.generatedCodeLine] = this.rawSegments = [];
      }

      this.generatedCodeColumn = 0;
    }

    this.generatedCodeColumn += lines[lines.length - 1].length;
  }

}

const n = '\n';
const warned = {
  insertLeft: false,
  insertRight: false,
  storeName: false
};

class MagicString {
  constructor(string, options = {}) {
    const chunk = new Chunk(0, string.length, string);
    Object.defineProperties(this, {
      original: {
        writable: true,
        value: string
      },
      outro: {
        writable: true,
        value: ''
      },
      intro: {
        writable: true,
        value: ''
      },
      firstChunk: {
        writable: true,
        value: chunk
      },
      lastChunk: {
        writable: true,
        value: chunk
      },
      lastSearchedChunk: {
        writable: true,
        value: chunk
      },
      byStart: {
        writable: true,
        value: {}
      },
      byEnd: {
        writable: true,
        value: {}
      },
      filename: {
        writable: true,
        value: options.filename
      },
      indentExclusionRanges: {
        writable: true,
        value: options.indentExclusionRanges
      },
      sourcemapLocations: {
        writable: true,
        value: new BitSet()
      },
      storedNames: {
        writable: true,
        value: {}
      },
      indentStr: {
        writable: true,
        value: undefined
      }
    });
    this.byStart[0] = chunk;
    this.byEnd[string.length] = chunk;
  }

  addSourcemapLocation(char) {
    this.sourcemapLocations.add(char);
  }

  append(content) {
    if (typeof content !== 'string') throw new TypeError('outro content must be a string');
    this.outro += content;
    return this;
  }

  appendLeft(index, content) {
    if (typeof content !== 'string') throw new TypeError('inserted content must be a string');

    this._split(index);

    const chunk = this.byEnd[index];

    if (chunk) {
      chunk.appendLeft(content);
    } else {
      this.intro += content;
    }

    return this;
  }

  appendRight(index, content) {
    if (typeof content !== 'string') throw new TypeError('inserted content must be a string');

    this._split(index);

    const chunk = this.byStart[index];

    if (chunk) {
      chunk.appendRight(content);
    } else {
      this.outro += content;
    }

    return this;
  }

  clone() {
    const cloned = new MagicString(this.original, {
      filename: this.filename
    });
    let originalChunk = this.firstChunk;
    let clonedChunk = cloned.firstChunk = cloned.lastSearchedChunk = originalChunk.clone();

    while (originalChunk) {
      cloned.byStart[clonedChunk.start] = clonedChunk;
      cloned.byEnd[clonedChunk.end] = clonedChunk;
      const nextOriginalChunk = originalChunk.next;
      const nextClonedChunk = nextOriginalChunk && nextOriginalChunk.clone();

      if (nextClonedChunk) {
        clonedChunk.next = nextClonedChunk;
        nextClonedChunk.previous = clonedChunk;
        clonedChunk = nextClonedChunk;
      }

      originalChunk = nextOriginalChunk;
    }

    cloned.lastChunk = clonedChunk;

    if (this.indentExclusionRanges) {
      cloned.indentExclusionRanges = this.indentExclusionRanges.slice();
    }

    cloned.sourcemapLocations = new BitSet(this.sourcemapLocations);
    cloned.intro = this.intro;
    cloned.outro = this.outro;
    return cloned;
  }

  generateDecodedMap(options) {
    options = options || {};
    const sourceIndex = 0;
    const names = Object.keys(this.storedNames);
    const mappings = new Mappings(options.hires);
    const locate = getLocator(this.original);

    if (this.intro) {
      mappings.advance(this.intro);
    }

    this.firstChunk.eachNext(chunk => {
      const loc = locate(chunk.start);
      if (chunk.intro.length) mappings.advance(chunk.intro);

      if (chunk.edited) {
        mappings.addEdit(sourceIndex, chunk.content, loc, chunk.storeName ? names.indexOf(chunk.original) : -1);
      } else {
        mappings.addUneditedChunk(sourceIndex, chunk, this.original, loc, this.sourcemapLocations);
      }

      if (chunk.outro.length) mappings.advance(chunk.outro);
    });
    return {
      file: options.file ? options.file.split(/[/\\]/).pop() : null,
      sources: [options.source ? getRelativePath(options.file || '', options.source) : null],
      sourcesContent: options.includeContent ? [this.original] : [null],
      names,
      mappings: mappings.raw
    };
  }

  generateMap(options) {
    return new SourceMap$1(this.generateDecodedMap(options));
  }

  _ensureindentStr() {
    if (this.indentStr === undefined) {
      this.indentStr = guessIndent(this.original);
    }
  }

  _getRawIndentString() {
    this._ensureindentStr();

    return this.indentStr;
  }

  getIndentString() {
    this._ensureindentStr();

    return this.indentStr === null ? '\t' : this.indentStr;
  }

  indent(indentStr, options) {
    const pattern = /^[^\r\n]/gm;

    if (isObject(indentStr)) {
      options = indentStr;
      indentStr = undefined;
    }

    if (indentStr === undefined) {
      this._ensureindentStr();

      indentStr = this.indentStr || '\t';
    }

    if (indentStr === '') return this; // noop

    options = options || {}; // Process exclusion ranges

    const isExcluded = {};

    if (options.exclude) {
      const exclusions = typeof options.exclude[0] === 'number' ? [options.exclude] : options.exclude;
      exclusions.forEach(exclusion => {
        for (let i = exclusion[0]; i < exclusion[1]; i += 1) {
          isExcluded[i] = true;
        }
      });
    }

    let shouldIndentNextCharacter = options.indentStart !== false;

    const replacer = match => {
      if (shouldIndentNextCharacter) return `${indentStr}${match}`;
      shouldIndentNextCharacter = true;
      return match;
    };

    this.intro = this.intro.replace(pattern, replacer);
    let charIndex = 0;
    let chunk = this.firstChunk;

    while (chunk) {
      const end = chunk.end;

      if (chunk.edited) {
        if (!isExcluded[charIndex]) {
          chunk.content = chunk.content.replace(pattern, replacer);

          if (chunk.content.length) {
            shouldIndentNextCharacter = chunk.content[chunk.content.length - 1] === '\n';
          }
        }
      } else {
        charIndex = chunk.start;

        while (charIndex < end) {
          if (!isExcluded[charIndex]) {
            const char = this.original[charIndex];

            if (char === '\n') {
              shouldIndentNextCharacter = true;
            } else if (char !== '\r' && shouldIndentNextCharacter) {
              shouldIndentNextCharacter = false;

              if (charIndex === chunk.start) {
                chunk.prependRight(indentStr);
              } else {
                this._splitChunk(chunk, charIndex);

                chunk = chunk.next;
                chunk.prependRight(indentStr);
              }
            }
          }

          charIndex += 1;
        }
      }

      charIndex = chunk.end;
      chunk = chunk.next;
    }

    this.outro = this.outro.replace(pattern, replacer);
    return this;
  }

  insert() {
    throw new Error('magicString.insert(...) is deprecated. Use prependRight(...) or appendLeft(...)');
  }

  insertLeft(index, content) {
    if (!warned.insertLeft) {
      console.warn('magicString.insertLeft(...) is deprecated. Use magicString.appendLeft(...) instead'); // eslint-disable-line no-console

      warned.insertLeft = true;
    }

    return this.appendLeft(index, content);
  }

  insertRight(index, content) {
    if (!warned.insertRight) {
      console.warn('magicString.insertRight(...) is deprecated. Use magicString.prependRight(...) instead'); // eslint-disable-line no-console

      warned.insertRight = true;
    }

    return this.prependRight(index, content);
  }

  move(start, end, index) {
    if (index >= start && index <= end) throw new Error('Cannot move a selection inside itself');

    this._split(start);

    this._split(end);

    this._split(index);

    const first = this.byStart[start];
    const last = this.byEnd[end];
    const oldLeft = first.previous;
    const oldRight = last.next;
    const newRight = this.byStart[index];
    if (!newRight && last === this.lastChunk) return this;
    const newLeft = newRight ? newRight.previous : this.lastChunk;
    if (oldLeft) oldLeft.next = oldRight;
    if (oldRight) oldRight.previous = oldLeft;
    if (newLeft) newLeft.next = first;
    if (newRight) newRight.previous = last;
    if (!first.previous) this.firstChunk = last.next;

    if (!last.next) {
      this.lastChunk = first.previous;
      this.lastChunk.next = null;
    }

    first.previous = newLeft;
    last.next = newRight || null;
    if (!newLeft) this.firstChunk = first;
    if (!newRight) this.lastChunk = last;
    return this;
  }

  overwrite(start, end, content, options) {
    options = options || {};
    return this.update(start, end, content, { ...options,
      overwrite: !options.contentOnly
    });
  }

  update(start, end, content, options) {
    if (typeof content !== 'string') throw new TypeError('replacement content must be a string');

    while (start < 0) start += this.original.length;

    while (end < 0) end += this.original.length;

    if (end > this.original.length) throw new Error('end is out of bounds');
    if (start === end) throw new Error('Cannot overwrite a zero-length range – use appendLeft or prependRight instead');

    this._split(start);

    this._split(end);

    if (options === true) {
      if (!warned.storeName) {
        console.warn('The final argument to magicString.overwrite(...) should be an options object. See https://github.com/rich-harris/magic-string'); // eslint-disable-line no-console

        warned.storeName = true;
      }

      options = {
        storeName: true
      };
    }

    const storeName = options !== undefined ? options.storeName : false;
    const overwrite = options !== undefined ? options.overwrite : false;

    if (storeName) {
      const original = this.original.slice(start, end);
      Object.defineProperty(this.storedNames, original, {
        writable: true,
        value: true,
        enumerable: true
      });
    }

    const first = this.byStart[start];
    const last = this.byEnd[end];

    if (first) {
      let chunk = first;

      while (chunk !== last) {
        if (chunk.next !== this.byStart[chunk.end]) {
          throw new Error('Cannot overwrite across a split point');
        }

        chunk = chunk.next;
        chunk.edit('', false);
      }

      first.edit(content, storeName, !overwrite);
    } else {
      // must be inserting at the end
      const newChunk = new Chunk(start, end, '').edit(content, storeName); // TODO last chunk in the array may not be the last chunk, if it's moved...

      last.next = newChunk;
      newChunk.previous = last;
    }

    return this;
  }

  prepend(content) {
    if (typeof content !== 'string') throw new TypeError('outro content must be a string');
    this.intro = content + this.intro;
    return this;
  }

  prependLeft(index, content) {
    if (typeof content !== 'string') throw new TypeError('inserted content must be a string');

    this._split(index);

    const chunk = this.byEnd[index];

    if (chunk) {
      chunk.prependLeft(content);
    } else {
      this.intro = content + this.intro;
    }

    return this;
  }

  prependRight(index, content) {
    if (typeof content !== 'string') throw new TypeError('inserted content must be a string');

    this._split(index);

    const chunk = this.byStart[index];

    if (chunk) {
      chunk.prependRight(content);
    } else {
      this.outro = content + this.outro;
    }

    return this;
  }

  remove(start, end) {
    while (start < 0) start += this.original.length;

    while (end < 0) end += this.original.length;

    if (start === end) return this;
    if (start < 0 || end > this.original.length) throw new Error('Character is out of bounds');
    if (start > end) throw new Error('end must be greater than start');

    this._split(start);

    this._split(end);

    let chunk = this.byStart[start];

    while (chunk) {
      chunk.intro = '';
      chunk.outro = '';
      chunk.edit('');
      chunk = end > chunk.end ? this.byStart[chunk.end] : null;
    }

    return this;
  }

  lastChar() {
    if (this.outro.length) return this.outro[this.outro.length - 1];
    let chunk = this.lastChunk;

    do {
      if (chunk.outro.length) return chunk.outro[chunk.outro.length - 1];
      if (chunk.content.length) return chunk.content[chunk.content.length - 1];
      if (chunk.intro.length) return chunk.intro[chunk.intro.length - 1];
    } while (chunk = chunk.previous);

    if (this.intro.length) return this.intro[this.intro.length - 1];
    return '';
  }

  lastLine() {
    let lineIndex = this.outro.lastIndexOf(n);
    if (lineIndex !== -1) return this.outro.substr(lineIndex + 1);
    let lineStr = this.outro;
    let chunk = this.lastChunk;

    do {
      if (chunk.outro.length > 0) {
        lineIndex = chunk.outro.lastIndexOf(n);
        if (lineIndex !== -1) return chunk.outro.substr(lineIndex + 1) + lineStr;
        lineStr = chunk.outro + lineStr;
      }

      if (chunk.content.length > 0) {
        lineIndex = chunk.content.lastIndexOf(n);
        if (lineIndex !== -1) return chunk.content.substr(lineIndex + 1) + lineStr;
        lineStr = chunk.content + lineStr;
      }

      if (chunk.intro.length > 0) {
        lineIndex = chunk.intro.lastIndexOf(n);
        if (lineIndex !== -1) return chunk.intro.substr(lineIndex + 1) + lineStr;
        lineStr = chunk.intro + lineStr;
      }
    } while (chunk = chunk.previous);

    lineIndex = this.intro.lastIndexOf(n);
    if (lineIndex !== -1) return this.intro.substr(lineIndex + 1) + lineStr;
    return this.intro + lineStr;
  }

  slice(start = 0, end = this.original.length) {
    while (start < 0) start += this.original.length;

    while (end < 0) end += this.original.length;

    let result = ''; // find start chunk

    let chunk = this.firstChunk;

    while (chunk && (chunk.start > start || chunk.end <= start)) {
      // found end chunk before start
      if (chunk.start < end && chunk.end >= end) {
        return result;
      }

      chunk = chunk.next;
    }

    if (chunk && chunk.edited && chunk.start !== start) throw new Error(`Cannot use replaced character ${start} as slice start anchor.`);
    const startChunk = chunk;

    while (chunk) {
      if (chunk.intro && (startChunk !== chunk || chunk.start === start)) {
        result += chunk.intro;
      }

      const containsEnd = chunk.start < end && chunk.end >= end;
      if (containsEnd && chunk.edited && chunk.end !== end) throw new Error(`Cannot use replaced character ${end} as slice end anchor.`);
      const sliceStart = startChunk === chunk ? start - chunk.start : 0;
      const sliceEnd = containsEnd ? chunk.content.length + end - chunk.end : chunk.content.length;
      result += chunk.content.slice(sliceStart, sliceEnd);

      if (chunk.outro && (!containsEnd || chunk.end === end)) {
        result += chunk.outro;
      }

      if (containsEnd) {
        break;
      }

      chunk = chunk.next;
    }

    return result;
  } // TODO deprecate this? not really very useful


  snip(start, end) {
    const clone = this.clone();
    clone.remove(0, start);
    clone.remove(end, clone.original.length);
    return clone;
  }

  _split(index) {
    if (this.byStart[index] || this.byEnd[index]) return;
    let chunk = this.lastSearchedChunk;
    const searchForward = index > chunk.end;

    while (chunk) {
      if (chunk.contains(index)) return this._splitChunk(chunk, index);
      chunk = searchForward ? this.byStart[chunk.end] : this.byEnd[chunk.start];
    }
  }

  _splitChunk(chunk, index) {
    if (chunk.edited && chunk.content.length) {
      // zero-length edited chunks are a special case (overlapping replacements)
      const loc = getLocator(this.original)(index);
      throw new Error(`Cannot split a chunk that has already been edited (${loc.line}:${loc.column} – "${chunk.original}")`);
    }

    const newChunk = chunk.split(index);
    this.byEnd[index] = chunk;
    this.byStart[index] = newChunk;
    this.byEnd[newChunk.end] = newChunk;
    if (chunk === this.lastChunk) this.lastChunk = newChunk;
    this.lastSearchedChunk = chunk;
    return true;
  }

  toString() {
    let str = this.intro;
    let chunk = this.firstChunk;

    while (chunk) {
      str += chunk.toString();
      chunk = chunk.next;
    }

    return str + this.outro;
  }

  isEmpty() {
    let chunk = this.firstChunk;

    do {
      if (chunk.intro.length && chunk.intro.trim() || chunk.content.length && chunk.content.trim() || chunk.outro.length && chunk.outro.trim()) return false;
    } while (chunk = chunk.next);

    return true;
  }

  length() {
    let chunk = this.firstChunk;
    let length = 0;

    do {
      length += chunk.intro.length + chunk.content.length + chunk.outro.length;
    } while (chunk = chunk.next);

    return length;
  }

  trimLines() {
    return this.trim('[\\r\\n]');
  }

  trim(charType) {
    return this.trimStart(charType).trimEnd(charType);
  }

  trimEndAborted(charType) {
    const rx = new RegExp((charType || '\\s') + '+$');
    this.outro = this.outro.replace(rx, '');
    if (this.outro.length) return true;
    let chunk = this.lastChunk;

    do {
      const end = chunk.end;
      const aborted = chunk.trimEnd(rx); // if chunk was trimmed, we have a new lastChunk

      if (chunk.end !== end) {
        if (this.lastChunk === chunk) {
          this.lastChunk = chunk.next;
        }

        this.byEnd[chunk.end] = chunk;
        this.byStart[chunk.next.start] = chunk.next;
        this.byEnd[chunk.next.end] = chunk.next;
      }

      if (aborted) return true;
      chunk = chunk.previous;
    } while (chunk);

    return false;
  }

  trimEnd(charType) {
    this.trimEndAborted(charType);
    return this;
  }

  trimStartAborted(charType) {
    const rx = new RegExp('^' + (charType || '\\s') + '+');
    this.intro = this.intro.replace(rx, '');
    if (this.intro.length) return true;
    let chunk = this.firstChunk;

    do {
      const end = chunk.end;
      const aborted = chunk.trimStart(rx);

      if (chunk.end !== end) {
        // special case...
        if (chunk === this.lastChunk) this.lastChunk = chunk.next;
        this.byEnd[chunk.end] = chunk;
        this.byStart[chunk.next.start] = chunk.next;
        this.byEnd[chunk.next.end] = chunk.next;
      }

      if (aborted) return true;
      chunk = chunk.next;
    } while (chunk);

    return false;
  }

  trimStart(charType) {
    this.trimStartAborted(charType);
    return this;
  }

  hasChanged() {
    return this.original !== this.toString();
  }

  _replaceRegexp(searchValue, replacement) {
    function getReplacement(match, str) {
      if (typeof replacement === 'string') {
        return replacement.replace(/\$(\$|&|\d+)/g, (_, i) => {
          // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace#specifying_a_string_as_a_parameter
          if (i === '$') return '$';
          if (i === '&') return match[0];
          const num = +i;
          if (num < match.length) return match[+i];
          return `$${i}`;
        });
      } else {
        return replacement(...match, match.index, str, match.groups);
      }
    }

    function matchAll(re, str) {
      let match;
      const matches = [];

      while (match = re.exec(str)) {
        matches.push(match);
      }

      return matches;
    }

    if (searchValue.global) {
      const matches = matchAll(searchValue, this.original);
      matches.forEach(match => {
        if (match.index != null) this.overwrite(match.index, match.index + match[0].length, getReplacement(match, this.original));
      });
    } else {
      const match = this.original.match(searchValue);
      if (match && match.index != null) this.overwrite(match.index, match.index + match[0].length, getReplacement(match, this.original));
    }

    return this;
  }

  _replaceString(string, replacement) {
    const {
      original
    } = this;
    const index = original.indexOf(string);

    if (index !== -1) {
      this.overwrite(index, index + string.length, replacement);
    }

    return this;
  }

  replace(searchValue, replacement) {
    if (typeof searchValue === 'string') {
      return this._replaceString(searchValue, replacement);
    }

    return this._replaceRegexp(searchValue, replacement);
  }

  _replaceAllString(string, replacement) {
    const {
      original
    } = this;
    const stringLength = string.length;

    for (let index = original.indexOf(string); index !== -1; index = original.indexOf(string, index + stringLength)) {
      this.overwrite(index, index + stringLength, replacement);
    }

    return this;
  }

  replaceAll(searchValue, replacement) {
    if (typeof searchValue === 'string') {
      return this._replaceAllString(searchValue, replacement);
    }

    if (!searchValue.global) {
      throw new TypeError('MagicString.prototype.replaceAll called with a non-global RegExp argument');
    }

    return this._replaceRegexp(searchValue, replacement);
  }

}

/**
 * Convert tabs indentation to two spaces.
 */

function tabs2Spaces(str) {
  return str.replace(/^\t+/, tabs => '  '.repeat(tabs.length));
}
/**
 * Generate an excerpt of the location in the source around the
 * specified position.
 */


function createCodeFrame(text, lineNum, columnNum, {
  before = 2,
  after = 3,
  colors = true,
  maxWidth = 0
} = {}) {
  const lines = text.split('\n');
  const start = Math.max(0, lineNum - before);
  const end = Math.min(lines.length, lineNum + after + 1); // Maximum space needed for line numbering in the current range.
  // Necessary when the amount of digits of the line numbering grows:
  //  999 | asdf
  // 1000 | asdjadfjsa

  const maxLineNum = String(end).length;
  const padding = ' '.repeat(maxLineNum); // Normalize all indentation (=tabs) to use 2 spaces. We need to
  // apply the difference to the marker position to move it back in
  // place.

  const spaceLines = [];
  let maxLineLen = 0;

  for (let i = start; i < end; i++) {
    const line = tabs2Spaces(lines[i]);
    spaceLines.push(line);
    if (line.length > maxLineLen) maxLineLen = line.length;
  }

  const activeLine = spaceLines[lineNum - start]; // Move marker into correct place by taking the amount of
  // normalized tabs into account

  const count = Math.max(0, activeLine.length - lines[lineNum].length + columnNum);
  const maxLensWidth = maxWidth - '> '.length - padding.length - ' | '.length;
  let left = 0;
  let right = maxLensWidth;

  if (maxWidth > 0) {
    const half = Math.floor(maxLensWidth / 2);
    let winLeft = count - half;

    if (winLeft > 0) {
      let winRight = count + half - 1;
      left = winLeft;
      right = winRight;

      if (winRight > maxLensWidth) {
        const offset = Math.min(0, winRight - maxLensWidth);
        left -= offset;
        right -= offset;
      }
    }
  }

  const sep = dim('|');
  let out = '';

  for (let i = 0; i < spaceLines.length; i++) {
    const line = spaceLines[i];
    const currentLine = dim((padding + (i + start + 1)).slice(-maxLineNum));
    let formatted = line;

    if (maxWidth > 0) {
      formatted = formatted.slice(left, Math.min(right, line.length));

      if (left > 0) {
        formatted = '…' + formatted;
      }

      if (line.length > right) {
        formatted += '…';
      }
    } // Line where the error occured


    if (i === lineNum - start) {
      out += red('>') + ` ${currentLine} ${sep} ${formatted}\n`;
      out += `  ${padding} ${sep} ${' '.repeat(count - left)}${bold(red('^'))}\n`;
    } else {
      out += `  ${currentLine} ${sep} ${formatted}\n`;
    }
  }

  return colors ? out : stripColors(out);
}

class ErrorWithLocation extends Error {
  constructor({
    message,
    filename,
    line,
    column
  }) {
    super(message);
    this.filename = filename;
    this.line = line;
    this.column = column;
  }

  async toCodeFrame() {
    if (!this.filename) throw new Error('filename missing in ErrorWithLocation');
    const originalCode = await node_fs.promises.readFile(this.filename, 'utf8');
    const frame = createCodeFrame(originalCode, this.line - 1, this.column || 0);
    const message = `${red(this.message)}

${blue(`${this.filename}:${this.line}${this.column === undefined ? '' : `:${this.column + 1}`}`)}

${frame}`;
    const modifiedError = new Error(message);
    modifiedError.stack = message;
    return modifiedError;
  }

}

// Matches the scheme of a URL, eg "http://"
const schemeRegex = /^[\w+.-]+:\/\//;
/**
 * Matches the parts of a URL:
 * 1. Scheme, including ":", guaranteed.
 * 2. User/password, including "@", optional.
 * 3. Host, guaranteed.
 * 4. Port, including ":", optional.
 * 5. Path, including "/", optional.
 */

const urlRegex = /^([\w+.-]+:)\/\/([^@]*@)?([^:/]*)(:\d+)?(\/[^#?]*)?/;

function isAbsoluteUrl(input) {
  return schemeRegex.test(input);
}

function isSchemeRelativeUrl(input) {
  return input.startsWith('//');
}

function isAbsolutePath(input) {
  return input.startsWith('/');
}

function parseAbsoluteUrl(input) {
  const match = urlRegex.exec(input);
  return {
    scheme: match[1],
    user: match[2] || '',
    host: match[3],
    port: match[4] || '',
    path: match[5] || '/',
    relativePath: false
  };
}

function parseUrl(input) {
  if (isSchemeRelativeUrl(input)) {
    const url = parseAbsoluteUrl('http:' + input);
    url.scheme = '';
    return url;
  }

  if (isAbsolutePath(input)) {
    const url = parseAbsoluteUrl('http://foo.com' + input);
    url.scheme = '';
    url.host = '';
    return url;
  }

  if (!isAbsoluteUrl(input)) {
    const url = parseAbsoluteUrl('http://foo.com/' + input);
    url.scheme = '';
    url.host = '';
    url.relativePath = true;
    return url;
  }

  return parseAbsoluteUrl(input);
}

function stripPathFilename(path) {
  // If a path ends with a parent directory "..", then it's a relative path with excess parent
  // paths. It's not a file, so we can't strip it.
  if (path.endsWith('/..')) return path;
  const index = path.lastIndexOf('/');
  return path.slice(0, index + 1);
}

function mergePaths(url, base) {
  // If we're not a relative path, then we're an absolute path, and it doesn't matter what base is.
  if (!url.relativePath) return;
  normalizePath(base); // If the path is just a "/", then it was an empty path to begin with (remember, we're a relative
  // path).

  if (url.path === '/') {
    url.path = base.path;
  } else {
    // Resolution happens relative to the base path's directory, not the file.
    url.path = stripPathFilename(base.path) + url.path;
  } // If the base path is absolute, then our path is now absolute too.


  url.relativePath = base.relativePath;
}
/**
 * The path can have empty directories "//", unneeded parents "foo/..", or current directory
 * "foo/.". We need to normalize to a standard representation.
 */


function normalizePath(url) {
  const {
    relativePath
  } = url;
  const pieces = url.path.split('/'); // We need to preserve the first piece always, so that we output a leading slash. The item at
  // pieces[0] is an empty string.

  let pointer = 1; // Positive is the number of real directories we've output, used for popping a parent directory.
  // Eg, "foo/bar/.." will have a positive 2, and we can decrement to be left with just "foo".

  let positive = 0; // We need to keep a trailing slash if we encounter an empty directory (eg, splitting "foo/" will
  // generate `["foo", ""]` pieces). And, if we pop a parent directory. But once we encounter a
  // real directory, we won't need to append, unless the other conditions happen again.

  let addTrailingSlash = false;

  for (let i = 1; i < pieces.length; i++) {
    const piece = pieces[i]; // An empty directory, could be a trailing slash, or just a double "//" in the path.

    if (!piece) {
      addTrailingSlash = true;
      continue;
    } // If we encounter a real directory, then we don't need to append anymore.


    addTrailingSlash = false; // A current directory, which we can always drop.

    if (piece === '.') continue; // A parent directory, we need to see if there are any real directories we can pop. Else, we
    // have an excess of parents, and we'll need to keep the "..".

    if (piece === '..') {
      if (positive) {
        addTrailingSlash = true;
        positive--;
        pointer--;
      } else if (relativePath) {
        // If we're in a relativePath, then we need to keep the excess parents. Else, in an absolute
        // URL, protocol relative URL, or an absolute path, we don't need to keep excess.
        pieces[pointer++] = piece;
      }

      continue;
    } // We've encountered a real directory. Move it to the next insertion pointer, which accounts for
    // any popped or dropped directories.


    pieces[pointer++] = piece;
    positive++;
  }

  let path = '';

  for (let i = 1; i < pointer; i++) {
    path += '/' + pieces[i];
  }

  if (!path || addTrailingSlash && !path.endsWith('/..')) {
    path += '/';
  }

  url.path = path;
}
/**
 * Attempts to resolve `input` URL/path relative to `base`.
 */


function resolve$2(input, base) {
  if (!input && !base) return '';
  const url = parseUrl(input); // If we have a base, and the input isn't already an absolute URL, then we need to merge.

  if (base && !url.scheme) {
    const baseUrl = parseUrl(base);
    url.scheme = baseUrl.scheme; // If there's no host, then we were just a path.

    if (!url.host || baseUrl.scheme === 'file:') {
      // The host, user, and port are joined, you can't copy one without the others.
      url.user = baseUrl.user;
      url.host = baseUrl.host;
      url.port = baseUrl.port;
    }

    mergePaths(url, baseUrl);
  }

  normalizePath(url); // If the input (and base, if there was one) are both relative, then we need to output a relative.

  if (url.relativePath) {
    // The first char is always a "/".
    const path = url.path.slice(1);
    if (!path) return '.'; // If base started with a leading ".", or there is no base and input started with a ".", then we
    // need to ensure that the relative path starts with a ".". We don't know if relative starts
    // with a "..", though, so check before prepending.

    const keepRelative = (base || input).startsWith('.');
    return !keepRelative || path.startsWith('.') ? path : './' + path;
  } // If there's no host (and no scheme/user/port), then we need to output an absolute path.


  if (!url.scheme && !url.host) return url.path; // We're outputting either an absolute URL, or a protocol relative one.

  return `${url.scheme}//${url.user}${url.host}${url.port}${url.path}`;
}

function resolve$1(input, base) {
  // The base is always treated as a directory, if it's not empty.
  // https://github.com/mozilla/source-map/blob/8cb3ee57/lib/util.js#L327
  // https://github.com/chromium/chromium/blob/da4adbb3/third_party/blink/renderer/devtools/front_end/sdk/SourceMap.js#L400-L401
  if (base && !base.endsWith('/')) base += '/';
  return resolve$2(input, base);
}
/**
 * Removes everything after the last "/", but leaves the slash.
 */


function stripFilename(path) {
  if (!path) return '';
  const index = path.lastIndexOf('/');
  return path.slice(0, index + 1);
}

const COLUMN = 0;

function maybeSort(mappings, owned) {
  const unsortedIndex = nextUnsortedSegmentLine(mappings, 0);
  if (unsortedIndex === mappings.length) return mappings; // If we own the array (meaning we parsed it from JSON), then we're free to directly mutate it. If
  // not, we do not want to modify the consumer's input array.

  if (!owned) mappings = mappings.slice();

  for (let i = unsortedIndex; i < mappings.length; i = nextUnsortedSegmentLine(mappings, i + 1)) {
    mappings[i] = sortSegments(mappings[i], owned);
  }

  return mappings;
}

function nextUnsortedSegmentLine(mappings, start) {
  for (let i = start; i < mappings.length; i++) {
    if (!isSorted(mappings[i])) return i;
  }

  return mappings.length;
}

function isSorted(line) {
  for (let j = 1; j < line.length; j++) {
    if (line[j][COLUMN] < line[j - 1][COLUMN]) {
      return false;
    }
  }

  return true;
}

function sortSegments(line, owned) {
  if (!owned) line = line.slice();
  return line.sort(sortComparator);
}

function sortComparator(a, b) {
  return a[COLUMN] - b[COLUMN];
}

let found = false;
/**
 * A binary search implementation that returns the index if a match is found.
 * If no match is found, then the left-index (the index associated with the item that comes just
 * before the desired index) is returned. To maintain proper sort order, a splice would happen at
 * the next index:
 *
 * ```js
 * const array = [1, 3];
 * const needle = 2;
 * const index = binarySearch(array, needle, (item, needle) => item - needle);
 *
 * assert.equal(index, 0);
 * array.splice(index + 1, 0, needle);
 * assert.deepEqual(array, [1, 2, 3]);
 * ```
 */

function binarySearch(haystack, needle, low, high) {
  while (low <= high) {
    const mid = low + (high - low >> 1);
    const cmp = haystack[mid][COLUMN] - needle;

    if (cmp === 0) {
      found = true;
      return mid;
    }

    if (cmp < 0) {
      low = mid + 1;
    } else {
      high = mid - 1;
    }
  }

  found = false;
  return low - 1;
}

function upperBound(haystack, needle, index) {
  for (let i = index + 1; i < haystack.length; index = i++) {
    if (haystack[i][COLUMN] !== needle) break;
  }

  return index;
}

function lowerBound(haystack, needle, index) {
  for (let i = index - 1; i >= 0; index = i--) {
    if (haystack[i][COLUMN] !== needle) break;
  }

  return index;
}

function memoizedState() {
  return {
    lastKey: -1,
    lastNeedle: -1,
    lastIndex: -1
  };
}
/**
 * This overly complicated beast is just to record the last tested line/column and the resulting
 * index, allowing us to skip a few tests if mappings are monotonically increasing.
 */


function memoizedBinarySearch(haystack, needle, state, key) {
  const {
    lastKey,
    lastNeedle,
    lastIndex
  } = state;
  let low = 0;
  let high = haystack.length - 1;

  if (key === lastKey) {
    if (needle === lastNeedle) {
      found = lastIndex !== -1 && haystack[lastIndex][COLUMN] === needle;
      return lastIndex;
    }

    if (needle >= lastNeedle) {
      // lastIndex may be -1 if the previous needle was not found.
      low = lastIndex === -1 ? 0 : lastIndex;
    } else {
      high = lastIndex;
    }
  }

  state.lastKey = key;
  state.lastNeedle = needle;
  return state.lastIndex = binarySearch(haystack, needle, low, high);
} // Rebuilds the original source files, with mappings that are ordered by source line/column instead
const LEAST_UPPER_BOUND = -1;
const GREATEST_LOWER_BOUND = 1;
/**
 * Returns the decoded (array of lines of segments) form of the SourceMap's mappings field.
 */

let decodedMappings;
/**
 * A low-level API to find the segment associated with a generated line/column (think, from a
 * stack trace). Line and column here are 0-based, unlike `originalPositionFor`.
 */

let traceSegment;

class TraceMap {
  constructor(map, mapUrl) {
    const isString = typeof map === 'string';
    if (!isString && map._decodedMemo) return map;
    const parsed = isString ? JSON.parse(map) : map;
    const {
      version,
      file,
      names,
      sourceRoot,
      sources,
      sourcesContent
    } = parsed;
    this.version = version;
    this.file = file;
    this.names = names;
    this.sourceRoot = sourceRoot;
    this.sources = sources;
    this.sourcesContent = sourcesContent;
    const from = resolve$1(sourceRoot || '', stripFilename(mapUrl));
    this.resolvedSources = sources.map(s => resolve$1(s || '', from));
    const {
      mappings
    } = parsed;

    if (typeof mappings === 'string') {
      this._encoded = mappings;
      this._decoded = undefined;
    } else {
      this._encoded = undefined;
      this._decoded = maybeSort(mappings, isString);
    }

    this._decodedMemo = memoizedState();
    this._bySources = undefined;
    this._bySourceMemos = undefined;
  }

}

(() => {

  decodedMappings = map => {
    return map._decoded || (map._decoded = decode(map._encoded));
  };

  traceSegment = (map, line, column) => {
    const decoded = decodedMappings(map); // It's common for parent source maps to have pointers to lines that have no
    // mapping (like a "//# sourceMappingURL=") at the end of the child file.

    if (line >= decoded.length) return null;
    return traceSegmentInternal(decoded[line], map._decodedMemo, line, column, GREATEST_LOWER_BOUND);
  };
})();

function traceSegmentInternal(segments, memo, line, column, bias) {
  let index = memoizedBinarySearch(segments, column, memo, line);

  if (found) {
    index = (bias === LEAST_UPPER_BOUND ? upperBound : lowerBound)(segments, column, index);
  } else if (bias === LEAST_UPPER_BOUND) index++;

  if (index === -1 || index === segments.length) return null;
  return segments[index];
}

/**
 * Gets the index associated with `key` in the backing array, if it is already present.
 */
let get;
/**
 * Puts `key` into the backing array, if it is not already present. Returns
 * the index of the `key` in the backing array.
 */

let put;
/**
 * SetArray acts like a `Set` (allowing only one occurrence of a string `key`), but provides the
 * index of the `key` in the backing array.
 *
 * This is designed to allow synchronizing a second array with the contents of the backing array,
 * like how in a sourcemap `sourcesContent[i]` is the source content associated with `source[i]`,
 * and there are never duplicates.
 */

class SetArray {
  constructor() {
    this._indexes = {
      __proto__: null
    };
    this.array = [];
  }

}

(() => {
  get = (strarr, key) => strarr._indexes[key];

  put = (strarr, key) => {
    // The key may or may not be present. If it is present, it's a number.
    const index = get(strarr, key);
    if (index !== undefined) return index;
    const {
      array,
      _indexes: indexes
    } = strarr;
    return indexes[key] = array.push(key) - 1;
  };
})();

/**
 * A low-level API to associate a generated position with an original source position. Line and
 * column here are 0-based, unlike `addMapping`.
 */

let addSegment;
/**
 * Adds/removes the content of the source file to the source map.
 */

let setSourceContent;
/**
 * Returns a sourcemap object (with decoded mappings) suitable for passing to a library that expects
 * a sourcemap, or to JSON.stringify.
 */

let decodedMap;
/**
 * Returns a sourcemap object (with encoded mappings) suitable for passing to a library that expects
 * a sourcemap, or to JSON.stringify.
 */

let encodedMap;
/**
 * Provides the state to generate a sourcemap.
 */

class GenMapping {
  constructor({
    file,
    sourceRoot
  } = {}) {
    this._names = new SetArray();
    this._sources = new SetArray();
    this._sourcesContent = [];
    this._mappings = [];
    this.file = file;
    this.sourceRoot = sourceRoot;
  }

}

(() => {
  addSegment = (map, genLine, genColumn, source, sourceLine, sourceColumn, name) => {
    const {
      _mappings: mappings,
      _sources: sources,
      _sourcesContent: sourcesContent,
      _names: names
    } = map;
    const line = getLine(mappings, genLine);

    if (source == null) {
      const seg = [genColumn];
      const index = getColumnIndex(line, genColumn, seg);
      return insert(line, index, seg);
    }

    const sourcesIndex = put(sources, source);
    const seg = name ? [genColumn, sourcesIndex, sourceLine, sourceColumn, put(names, name)] : [genColumn, sourcesIndex, sourceLine, sourceColumn];
    const index = getColumnIndex(line, genColumn, seg);
    if (sourcesIndex === sourcesContent.length) sourcesContent[sourcesIndex] = null;
    insert(line, index, seg);
  };

  setSourceContent = (map, source, content) => {
    const {
      _sources: sources,
      _sourcesContent: sourcesContent
    } = map;
    sourcesContent[put(sources, source)] = content;
  };

  decodedMap = map => {
    const {
      file,
      sourceRoot,
      _mappings: mappings,
      _sources: sources,
      _sourcesContent: sourcesContent,
      _names: names
    } = map;
    return {
      version: 3,
      file,
      names: names.array,
      sourceRoot: sourceRoot || undefined,
      sources: sources.array,
      sourcesContent,
      mappings
    };
  };

  encodedMap = map => {
    const decoded = decodedMap(map);
    return Object.assign(Object.assign({}, decoded), {
      mappings: encode(decoded.mappings)
    });
  };
})();

function getLine(mappings, index) {
  for (let i = mappings.length; i <= index; i++) {
    mappings[i] = [];
  }

  return mappings[index];
}

function getColumnIndex(line, column, seg) {
  let index = line.length;

  for (let i = index - 1; i >= 0; i--, index--) {
    const current = line[i];
    const col = current[0];
    if (col > column) continue;
    if (col < column) break;
    const cmp = compare(current, seg);
    if (cmp === 0) return index;
    if (cmp < 0) break;
  }

  return index;
}

function compare(a, b) {
  let cmp = compareNum(a.length, b.length);
  if (cmp !== 0) return cmp; // We've already checked genColumn

  if (a.length === 1) return 0;
  cmp = compareNum(a[1], b[1]);
  if (cmp !== 0) return cmp;
  cmp = compareNum(a[2], b[2]);
  if (cmp !== 0) return cmp;
  cmp = compareNum(a[3], b[3]);
  if (cmp !== 0) return cmp;
  if (a.length === 4) return 0;
  return compareNum(a[4], b[4]);
}

function compareNum(a, b) {
  return a - b;
}

function insert(array, index, value) {
  if (index === -1) return;

  for (let i = array.length; i > index; i--) {
    array[i] = array[i - 1];
  }

  array[index] = value;
}

const SOURCELESS_MAPPING = {
  source: null,
  column: null,
  line: null,
  name: null,
  content: null
};
const EMPTY_SOURCES = [];

function Source(map, sources, source, content) {
  return {
    map,
    sources,
    source,
    content
  };
}
/**
 * MapSource represents a single sourcemap, with the ability to trace mappings into its child nodes
 * (which may themselves be SourceMapTrees).
 */


function MapSource(map, sources) {
  return Source(map, sources, '', null);
}
/**
 * A "leaf" node in the sourcemap tree, representing an original, unmodified source file. Recursive
 * segment tracing ends at the `OriginalSource`.
 */


function OriginalSource(source, content) {
  return Source(null, EMPTY_SOURCES, source, content);
}
/**
 * traceMappings is only called on the root level SourceMapTree, and begins the process of
 * resolving each mapping in terms of the original source files.
 */


function traceMappings(tree) {
  const gen = new GenMapping({
    file: tree.map.file
  });
  const {
    sources: rootSources,
    map
  } = tree;
  const rootNames = map.names;
  const rootMappings = decodedMappings(map);

  for (let i = 0; i < rootMappings.length; i++) {
    const segments = rootMappings[i];
    let lastSource = null;
    let lastSourceLine = null;
    let lastSourceColumn = null;

    for (let j = 0; j < segments.length; j++) {
      const segment = segments[j];
      const genCol = segment[0];
      let traced = SOURCELESS_MAPPING; // 1-length segments only move the current generated column, there's no source information
      // to gather from it.

      if (segment.length !== 1) {
        const source = rootSources[segment[1]];
        traced = originalPositionFor(source, segment[2], segment[3], segment.length === 5 ? rootNames[segment[4]] : ''); // If the trace is invalid, then the trace ran into a sourcemap that doesn't contain a
        // respective segment into an original source.

        if (traced == null) continue;
      } // So we traced a segment down into its original source file. Now push a
      // new segment pointing to this location.


      const {
        column,
        line,
        name,
        content,
        source
      } = traced;

      if (line === lastSourceLine && column === lastSourceColumn && source === lastSource) {
        continue;
      }

      lastSourceLine = line;
      lastSourceColumn = column;
      lastSource = source; // Sigh, TypeScript can't figure out source/line/column are either all null, or all non-null...

      addSegment(gen, i, genCol, source, line, column, name);
      if (content != null) setSourceContent(gen, source, content);
    }
  }

  return gen;
}
/**
 * originalPositionFor is only called on children SourceMapTrees. It recurses down into its own
 * child SourceMapTrees, until we find the original source map.
 */


function originalPositionFor(source, line, column, name) {
  if (!source.map) {
    return {
      column,
      line,
      name,
      source: source.source,
      content: source.content
    };
  }

  const segment = traceSegment(source.map, line, column); // If we couldn't find a segment, then this doesn't exist in the sourcemap.

  if (segment == null) return null; // 1-length segments only move the current generated column, there's no source information
  // to gather from it.

  if (segment.length === 1) return SOURCELESS_MAPPING;
  return originalPositionFor(source.sources[segment[1]], segment[2], segment[3], segment.length === 5 ? source.map.names[segment[4]] : name);
}

function asArray(value) {
  if (Array.isArray(value)) return value;
  return [value];
}
/**
 * Recursively builds a tree structure out of sourcemap files, with each node
 * being either an `OriginalSource` "leaf" or a `SourceMapTree` composed of
 * `OriginalSource`s and `SourceMapTree`s.
 *
 * Every sourcemap is composed of a collection of source files and mappings
 * into locations of those source files. When we generate a `SourceMapTree` for
 * the sourcemap, we attempt to load each source file's own sourcemap. If it
 * does not have an associated sourcemap, it is considered an original,
 * unmodified source file.
 */


function buildSourceMapTree(input, loader) {
  const maps = asArray(input).map(m => new TraceMap(m, ''));
  const map = maps.pop();

  for (let i = 0; i < maps.length; i++) {
    if (maps[i].sources.length > 1) {
      throw new Error(`Transformation map ${i} must have exactly one source file.\n` + 'Did you specify these with the most recent transformation maps first?');
    }
  }

  let tree = build(map, loader, '', 0);

  for (let i = maps.length - 1; i >= 0; i--) {
    tree = MapSource(maps[i], [tree]);
  }

  return tree;
}

function build(map, loader, importer, importerDepth) {
  const {
    resolvedSources,
    sourcesContent
  } = map;
  const depth = importerDepth + 1;
  const children = resolvedSources.map((sourceFile, i) => {
    // The loading context gives the loader more information about why this file is being loaded
    // (eg, from which importer). It also allows the loader to override the location of the loaded
    // sourcemap/original source, or to override the content in the sourcesContent field if it's
    // an unmodified source file.
    const ctx = {
      importer,
      depth,
      source: sourceFile || '',
      content: undefined
    }; // Use the provided loader callback to retrieve the file's sourcemap.
    // TODO: We should eventually support async loading of sourcemap files.

    const sourceMap = loader(ctx.source, ctx);
    const {
      source,
      content
    } = ctx; // If there is a sourcemap, then we need to recurse into it to load its source files.

    if (sourceMap) return build(new TraceMap(sourceMap, source), loader, source, depth); // Else, it's an an unmodified source file.
    // The contents of this unmodified source file can be overridden via the loader context,
    // allowing it to be explicitly null or a string. If it remains undefined, we fall back to
    // the importing sourcemap's `sourcesContent` field.

    const sourceContent = content !== undefined ? content : sourcesContent ? sourcesContent[i] : null;
    return OriginalSource(source, sourceContent);
  });
  return MapSource(map, children);
}
/**
 * A SourceMap v3 compatible sourcemap, which only includes fields that were
 * provided to it.
 */


class SourceMap {
  constructor(map, options) {
    const out = options.decodedMappings ? decodedMap(map) : encodedMap(map);
    this.version = out.version; // SourceMap spec says this should be first.

    this.file = out.file;
    this.mappings = out.mappings;
    this.names = out.names;
    this.sourceRoot = out.sourceRoot;
    this.sources = out.sources;

    if (!options.excludeContent) {
      this.sourcesContent = out.sourcesContent;
    }
  }

  toString() {
    return JSON.stringify(this);
  }

}
/**
 * Traces through all the mappings in the root sourcemap, through the sources
 * (and their sourcemaps), all the way back to the original source location.
 *
 * `loader` will be called every time we encounter a source file. If it returns
 * a sourcemap, we will recurse into that sourcemap to continue the trace. If
 * it returns a falsey value, that source file is treated as an original,
 * unmodified source file.
 *
 * Pass `excludeContent` to exclude any self-containing source file content
 * from the output sourcemap.
 *
 * Pass `decodedMappings` to receive a SourceMap with decoded (instead of
 * VLQ encoded) mappings.
 */


function remapping(input, loader, options) {
  const opts = typeof options === 'object' ? options : {
    excludeContent: !!options,
    decodedMappings: false
  };
  const tree = buildSourceMapTree(input, loader);
  return new SourceMap(traceMappings(tree), opts);
}

// Copied from https://github.com/vitejs/vite/blob/d97b33a8cb9a72ed64244f239900a9a862b6ba68/packages/vite/src/node/utils.ts#L431
// Based on https://github.com/sveltejs/svelte/blob/abf11bb02b2afbd3e4cac509a0f70e318c306364/src/compiler/utils/mapped_code.ts#L221
const nullSourceMap = {
  names: [],
  sources: [],
  mappings: '',
  version: 3
};
const combineSourceMaps = (filename, _sourceMapList) => {
  const sourceMapList = _sourceMapList.map(map => {
    // eslint-disable-next-line @cloudfour/typescript-eslint/no-unnecessary-condition
    if (!map.sources) map.sources = [];
    return map;
  }).reverse();

  if (sourceMapList.every(m => m.sources.length === 0)) return { ...nullSourceMap
  };
  let mapIndex = 1;
  const useArrayInterface = !sourceMapList.slice(0, -1).some(m => m.sources.length !== 1);
  const map = useArrayInterface ? remapping(sourceMapList, () => null, true) : remapping(sourceMapList[0], sourcefile => sourcefile === filename && sourceMapList[mapIndex] ? sourceMapList[mapIndex++] : { ...nullSourceMap
  }, true);
  if (!map.file) delete map.file;
  return map;
};

/*
 * Refactored based on https://github.com/preactjs/wmr/blob/main/packages/wmr/src/lib/rollup-plugin-container.js
 * https://github.com/preactjs/wmr/blob/main/LICENSE
 * MIT License
 * Copyright (c) 2020 The Preact Authors
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/** Fast splice(x,1) when order doesn't matter (h/t Rich) */
const popIndex = (array, index) => {
  const tail = array.pop();
  if (index !== array.length) array[index] = tail;
};
/** Get a unique key for a (id,importer) resolve pair */


const identifierPair = (id, importer) => {
  if (importer) return `${id}\n${importer}`;
  return id;
};

const createPluginContainer = plugins => {
  const MODULES = new Map();
  let plugin;
  const parser = acorn.Parser;
  const ctx = {
    meta: {
      rollupVersion: '2.8.0',
      watchMode: true
    },

    parse(code, opts) {
      return parser.parse(code, {
        sourceType: 'module',
        ecmaVersion: 2020,
        locations: true,
        onComment: [],
        ...opts
      });
    },

    async resolve(id, importer, {
      skipSelf = false
    } = {}) {
      const skip = [];
      if (skipSelf && plugin) skip.push(plugin);
      let out = await container.resolveId(id, importer, skip);
      if (typeof out === 'string') out = {
        id: out
      };
      if (!out || !out.id) out = {
        id
      };

      if (/^\.\.?[/\\]/.test(out.id) || out.id === '.') {
        out.id = path.resolve('.', importer ? path.dirname(importer) : '.', out.id);
      }

      return out || false;
    },

    getModuleInfo(id) {
      let mod = MODULES.get(id);
      if (mod) return mod.info;
      mod = {
        info: {}
      };
      MODULES.set(id, mod);
      return mod.info;
    },

    emitFile() {
      throw new Error('emitFile is not implemented');
    },

    setAssetSource() {
      throw new Error('setAssetSource is not implemented');
    },

    getFileName() {
      throw new Error('getFileName is not implemented');
    },

    addWatchFile() {// No-op
    },

    getWatchFiles() {
      throw new Error('getWatchFiles is not implemented');
    },

    getModuleIds() {
      throw new Error('getModuleIds is not implemented');
    },

    warn(...args) {
      var _plugin;

      console.log(`[${(_plugin = plugin) == null ? void 0 : _plugin.name}]`, ...args);
    },

    error(error, pos) {
      var _plugin2;

      if (pos === undefined) {
        if (typeof error === 'string') throw new Error(error);
        throw error;
      }

      throw new ErrorWithLocation({
        message: `[${(_plugin2 = plugin) == null ? void 0 : _plugin2.name}] ${error}`,
        line: typeof pos === 'number' ? pos : pos.line,
        column: pos.column
      });
    }

  };
  const container = {
    ctx,

    async buildStart() {
      await Promise.all(getSortedPluginsByHook('buildStart', plugins).map(async plugin => {
        if (plugin.buildStart) {
          const f = 'handler' in plugin.buildStart ? plugin.buildStart.handler : plugin.buildStart;
          await f.call(ctx, {});
        }
      }));
    },

    async resolveId(id, importer, _skip) {
      const key = identifierPair(id, importer);
      const opts = {};

      for (const p of getSortedPluginsByHook('resolveId', plugins)) {
        if (!p.resolveId) continue;

        if (_skip) {
          if (_skip.includes(p)) continue;
          if (resolveSkips.has(p, key)) continue;
          resolveSkips.add(p, key);
        }

        plugin = p;
        let result;

        try {
          const f = 'handler' in p.resolveId ? p.resolveId.handler : p.resolveId;
          result = await f.call(ctx, id, importer, {
            isEntry: false
          });
        } finally {
          if (_skip) resolveSkips.delete(p, key);
        }

        if (!result) continue;

        if (typeof result === 'string') {
          id = result;
        } else {
          id = result.id;
          Object.assign(opts, result);
        } // ResolveId() is hookFirst - first non-null result is returned.


        break;
      }

      opts.id = id;
      return Object.keys(opts).length > 1 ? opts : id;
    },

    async transform(originalCode, id, inputMap) {
      let code = originalCode; // TODO: if any of the transforms is missing sourcemaps, then there should be no source maps emitted

      const sourceMaps = inputMap ? [typeof inputMap === 'string' ? JSON.parse(inputMap) : inputMap] : [];

      for (plugin of getSortedPluginsByHook('transform', plugins)) {
        if (!plugin.transform) continue;

        try {
          const f = 'handler' in plugin.transform ? plugin.transform.handler : plugin.transform;
          const result = await f.call(ctx, code, id);
          if (!result) continue;

          if (typeof result === 'object') {
            if (!result.code) continue;
            code = result.code;
            if (result.map) sourceMaps.push(typeof result.map === 'string' ? JSON.parse(result.map) : result.map);
          } else {
            code = result;
          }
        } catch (error) {
          if (error instanceof ErrorWithLocation) {
            if (!error.filename) error.filename = id; // If the error has a location,
            // apply the source maps to get the original location

            const line = error.line;
            const column = error.column || 0;

            if (sourceMaps.length > 0) {
              const {
                SourceMapConsumer
              } = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require('source-map')); });
              const consumer = await new SourceMapConsumer(combineSourceMaps(id, sourceMaps));
              const sourceLocation = consumer.originalPositionFor({
                line,
                column
              });
              consumer.destroy();

              if (sourceLocation.line !== null) {
                error.line = sourceLocation.line;
                error.column = sourceLocation.column === null ? undefined : sourceLocation.column;
              }

              error.filename = sourceLocation.source || id;
            }
          }

          throw error;
        }
      }

      return {
        code,
        map: combineSourceMaps(id, sourceMaps)
      };
    },

    async options() {
      for (plugin of getSortedPluginsByHook('options', plugins)) {
        // Since we don't have "input options", we just pass {}
        // This hook must be called for @rollup/plugin-babel
        if (!plugin.options) continue;
        const f = 'handler' in plugin.options ? plugin.options.handler : plugin.options;
        await f.call(ctx, {});
      }
    },

    async load(id) {
      for (plugin of getSortedPluginsByHook('load', plugins)) {
        if (!plugin.load) continue;
        const f = 'handler' in plugin.load ? plugin.load.handler : plugin.load;
        const result = await f.call(ctx, id);

        if (result) {
          return result;
        }
      }

      return null;
    }

  }; // Tracks recursive resolveId calls

  const resolveSkips = {
    skip: new Map(),

    has(plugin, key) {
      const skips = this.skip.get(plugin);
      return skips ? skips.includes(key) : false;
    },

    add(plugin, key) {
      const skips = this.skip.get(plugin);
      if (skips) skips.push(key);else this.skip.set(plugin, [key]);
    },

    delete(plugin, key) {
      const skips = this.skip.get(plugin);
      if (!skips) return;
      const i = skips.indexOf(key);
      if (i !== -1) popIndex(skips, i);
    }

  };
  return container;
};

const getSortedPluginsByHook = (hookName, plugins) => {
  const pre = [];
  const normal = [];
  const post = [];

  for (const plugin of plugins) {
    const hook = plugin[hookName];

    if (hook) {
      if (typeof hook === 'object') {
        if (hook.order === 'pre') {
          pre.push(plugin);
          continue;
        }

        if (hook.order === 'post') {
          post.push(plugin);
          continue;
        }
      }

      normal.push(plugin);
    }
  }

  return [...pre, ...normal, ...post];
};

// Copied from https://github.com/preactjs/wmr/blob/c03abcc36c26dc936af8701ab9031ddff44995a5/packages/wmr/src/plugins/resolve-extensions-plugin.js
const transformImports = async (code, id, map, {
  resolveImportMeta,
  resolveId,
  resolveDynamicImport
} = {}) => {
  let imports;

  try {
    // eslint-disable-next-line @cloudfour/typescript-eslint/await-thenable
    const parsed = await esModuleLexer.parse(code, id);
    imports = parsed[0];
  } catch (error) {
    if (!('idx' in error)) throw error;
    const linesUntilError = code.slice(0, error.idx).split('\n');
    const line = linesUntilError.length;
    const column = linesUntilError[linesUntilError.length - 1].length;
    const ext = path.extname(id);
    const suggestion = jsExts.test(ext) ? '' : ` Did you mean to add a transform plugin to support ${ext} files?`;
    const modifiedError = new ErrorWithLocation({
      message: `Error parsing module with es-module-lexer.${suggestion}`,
      line,
      column,
      filename: id
    });

    if (map) {
      const {
        SourceMapConsumer
      } = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require('source-map')); });
      const consumer = await new SourceMapConsumer(map);
      const sourceLocation = consumer.originalPositionFor({
        line,
        column
      });
      consumer.destroy();

      if (sourceLocation.line !== null) {
        modifiedError.line = sourceLocation.line;
        modifiedError.column = sourceLocation.column === null ? undefined : sourceLocation.column;
      }

      modifiedError.filename = sourceLocation.source || id;
    }

    throw modifiedError;
  }

  let out = '';
  let offset = 0; // Import specifiers are synchronously converted into placeholders.
  // Resolutions are async+parallel, held in a mapping to their placeholders.

  let resolveIds = 0;
  const toResolve = new Map(); // Get a [deduplicated] placeholder for a specifier and kick off resolution

  const field = (spec, resolver, a, b) => {
    const match = toResolve.get(spec);
    if (match) return match.placeholder;
    const placeholder = `%%_RESOLVE_#${++resolveIds}#_%%`;
    toResolve.set(spec, {
      placeholder,
      spec,
      p: resolver(a, b)
    });
    return placeholder;
  }; // Falls through to resolveId() if null, preserves spec if false


  const doResolveDynamicImport = async (spec, id) => {
    let f = resolveDynamicImport && (await resolveDynamicImport(spec, id));
    if ((f === null || f === undefined) && resolveId) f = await resolveId(spec, id);
    return f;
  };

  for (const item of imports) {
    // Skip items that were already processed by being wrapped in an import - eg `import(import.meta.url)`
    if (item.s < offset) continue;
    out += code.slice(offset, item.s);
    const isImportMeta = item.d === -2;
    const isDynamicImport = item.d > -1;

    if (isDynamicImport) {
      // Bugfix: `import(import.meta.url)` returns an invalid negative end_offset.
      // We detect that here and find the closing paren to estimate the offset.
      if (item.e < 0) {
        // @ts-expect-error it is not readonly
        item.e = code.indexOf(')', item.s);
      } // Dynamic import() has no statement_end, so we take the position following the closing paren:
      // @ts-expect-error it is not readonly


      item.se = code.indexOf(')', item.e) + 1;
    }

    let quote = '';
    const after = code.slice(item.e, item.se);
    let spec = code.slice(item.s, item.e);
    offset = item.se;

    if (isImportMeta) {
      // Check for *simple* property access immediately following `import.meta`:
      const r = /\s*\.\s*([$_a-z][\w$]*)/gi;
      r.lastIndex = offset;
      const match = r.exec(code);

      if (match && match.index === offset) {
        // Advance past it and append it to the "specifier":
        offset = r.lastIndex;
        spec += match[0]; // Resolve it:

        const property = match[1];

        if (resolveImportMeta) {
          spec = field(spec, resolveImportMeta, property);
        }
      }

      out += spec;
      continue;
    } // Dynamic import()


    if (isDynamicImport) {
      // Strip comments - these are usually Webpack magic comments.
      spec = spec.replace(/\/\*[\S\s]*\*\//g, '').replace(/^\s*\/\/.*$/gm, '').trim(); // For dynamic imports, spec is a JavaScript expression.
      // We need to try to convert it to a specifier, or bail if it's not static.

      quote = (/^\s*(["'`])/.exec(spec) || [])[1];

      if (!quote) {
        // Warn if we were supposed to pass an AST node to resolveDynamicImport(), which is not implemented.
        if (resolveDynamicImport) {
          console.warn(`Cannot resolve dynamic expression in import(${spec})`);
        }

        out += spec + after;
        continue;
      }

      spec = spec.replace(/^\s*(["'`])(.*)\1\s*$/g, '$2');

      if (resolveDynamicImport) {
        spec = field(spec, doResolveDynamicImport, spec, id);
        out += quote + spec + quote + after;
        continue;
      }
    }

    if (resolveId) {
      spec = field(spec, resolveId, spec, id);
    }

    out += quote + spec + quote + after;
  }

  out += code.slice(Math.max(0, offset)); // Wait for all resolutions to finish and map them to placeholders

  const mapping = new Map();
  await Promise.all([...toResolve.values()].map(async v => {
    mapping.set(v.placeholder, (await v.p) || v.spec);
  }));
  out = out.replace(/%%_RESOLVE_#\d+#_%%/g, s => mapping.get(s));
  return out;
};

const getResolveCacheKey = (spec, from) => `${spec}%%FROM%%${from}`; // Minimal version of https://github.com/preactjs/wmr/blob/main/packages/wmr/src/wmr-middleware.js


const jsMiddleware = async ({
  root,
  plugins,
  requestCache
}) => {
  /**
   * The resolve cache is used so that if something has already been resolved from a previous build,
   * the buildId from the previous build gets used rather than the current buildId.
   * That way, modules can get correctly deduped in the browser,
   * and syntax/transform errors will get thrown from the _first_ runJS/loadJS that they were imported from.
   */
  const resolveCache = new Map();

  const setInResolveCache = (spec, from, buildId, resolved) => resolveCache.set(getResolveCacheKey(spec, from), {
    buildId,
    resolved
  });

  const getFromResolveCache = (spec, from) => resolveCache.get(getResolveCacheKey(spec, from));

  const rollupPlugins = createPluginContainer(plugins);
  await rollupPlugins.options();
  await rollupPlugins.buildStart();
  return async (req, res, next) => {
    const buildId = req.query['build-id'] !== undefined && Number(req.query['build-id']);

    try {
      // Normalized path starting with slash
      const path$1 = path.posix.normalize(req.path);
      const params = new URLSearchParams(req.query);
      let id;
      let file;

      if (path$1.startsWith('/@npm/')) {
        id = path$1.slice(1); // Remove leading slash

        file = params.get('resolvedPath') || '';
      } else {
        // Remove leading slash, and convert slashes to os-specific slashes
        const osPath = path$1.slice(1).split(path.posix.sep).join(path.sep); // Absolute file path

        file = path.resolve(root, osPath); // Rollup-style Unix-normalized path "id":

        id = file.split(path.sep).join(path.posix.sep);
      }

      params.delete('import');
      params.delete('inline-code');
      params.delete('build-id'); // Remove trailing =
      // This is necessary for rollup-plugin-vue, which ads ?lang.ts at the end of the id,
      // so the file gets processed by other transformers

      const qs = params.toString().replace(/=$/, '');
      if (qs) id += `?${qs}`;
      res.setHeader('Content-Type', 'application/javascript;charset=utf-8');
      const resolved = await rollupPlugins.resolveId(id);
      const resolvedId = typeof resolved === 'object' ? resolved == null ? void 0 : resolved.id : resolved;
      let code;
      let map;

      if (typeof req.query['inline-code'] === 'string') {
        code = req.query['inline-code'];
        const injectedArgsCode = `if (window._pleasantestArgs) {
          import.meta.pleasantestArgs = [...window._pleasantestArgs]
        }`;
        const fileSrc = await node_fs.promises.readFile(file, 'utf8');
        const inlineStartIdx = fileSrc.indexOf(code);
        code = injectedArgsCode + code;

        if (inlineStartIdx !== -1) {
          const str = new MagicString(fileSrc);
          str.remove(0, inlineStartIdx);
          str.remove(inlineStartIdx + code.length, fileSrc.length); // Account for the injected import.meta.pleasantestArgs code in the source map

          str.prepend(injectedArgsCode);
          map = str.generateMap({
            hires: true,
            source: id,
            includeContent: true
          });
        }
      } else {
        const result = resolvedId && (await rollupPlugins.load(resolvedId));

        if (typeof result === 'object' && result != null && result.map) {
          map = result.map;
        }

        code = typeof result === 'object' ? result == null ? void 0 : result.code : result;
      }

      if (!code && code !== '') {
        // If it doesn't have a js-like extension,
        // and none of the rollup plugins provided a load hook for it
        // and it doesn't have the ?import param (added for non-JS assets that can be imported into JS, like css)
        // Then treat it as a static asset
        if (!jsExts.test(resolvedId || req.path) && req.query.import === undefined) {
          next();
          return;
        }

        code = await node_fs.promises.readFile(file, 'utf8');
      }

      const transformResult = await rollupPlugins.transform(code, id, map);
      requestCache.set(id, transformResult);
      code = transformResult.code; // Normalize import paths
      // Resolve all the imports and replace them, and inline the resulting resolved paths
      // This makes different ways of importing the same path (e.g. extensionless imports, etc.)
      // all dedupe to the same module so it is only executed once

      code = await transformImports(code, id, map, {
        async resolveId(spec) {
          const addBuildId = specifier => {
            const delimiter = /\?/.test(specifier) ? '&' : '?';
            return `${specifier}${delimiter}build-id=${localBuildId}`;
          }; // Default to the buildId corresponding to this module
          // But for any module which has previously been imported from another buildId,
          // Use the previous buildId (for module deduplication in the browser)


          let localBuildId = buildId;
          if (/^(data:|https?:|\/\/)/.test(spec)) return spec;
          const cached = getFromResolveCache(spec, file);
          let resolved;

          if (cached) {
            resolved = cached.resolved;
            localBuildId = cached.buildId;
          } else {
            resolved = await rollupPlugins.resolveId(spec, file);
            if (resolved && buildId) setInResolveCache(spec, file, buildId, typeof resolved === 'object' ? resolved : {
              id: resolved
            });
          }

          if (resolved) {
            spec = typeof resolved === 'object' ? resolved.id : resolved;
            if (spec.startsWith('@npm/')) return addBuildId(`/${spec}`);

            if (/^(\/|\\|[a-z]:\\)/i.test(spec)) {
              // Change FS-absolute paths to relative
              spec = path.relative(path.dirname(file), spec).split(path.sep).join(path.posix.sep);
              if (!/^\.?\.?\//.test(spec)) spec = `./${spec}`;
            }

            if (typeof resolved === 'object' && resolved.external) {
              if (/^(data|https?):/.test(spec)) return spec;
              spec = path.relative(root, spec).split(path.sep).join(path.posix.sep);
              if (!/^(\/|[\w-]+:)/.test(spec)) spec = `/${spec}`;
              return addBuildId(spec);
            }
          } // If it wasn't resolved, and doesn't have a js-like extension
          // add the ?import query param to make it clear
          // that the request needs to end up as JS that can be imported


          if (!jsExts.test(spec)) {
            // If there is already a query parameter, add &import
            const delimiter = /\?/.test(spec) ? '&' : '?';
            return addBuildId(`${spec}${delimiter}import`);
          }

          return addBuildId(spec);
        }

      });

      if (!code) {
        next();
        return;
      }

      res.writeHead(200, {
        'Content-Length': Buffer.byteLength(code, 'utf8')
      });
      res.end(code);
    } catch (error) {
      if (buildId) {
        rejectBuild(Number(buildId), error instanceof ErrorWithLocation ? await error.toCodeFrame().catch(() => error) : error);
        res.statusCode = 500;
        return res.end();
      }

      next(error);
    }
  };
};

/**
 * This middleware handles static assets that are requested
 */
const staticMiddleware = ({
  root
}) => async (req, res, next) => {
  try {
    const absPath = path.resolve(root, ...req.path.split(path.posix.sep));

    if (!absPath.startsWith(root)) {
      next();
      return;
    }

    const stats = await node_fs.promises.stat(absPath).catch(() => {});

    if (!(stats != null && stats.isFile())) {
      next();
      return;
    }

    const headers = {
      'Content-Type': mime__default.getType(absPath) || ''
    };
    res.writeHead(200, headers);
    node_fs.createReadStream(absPath).pipe(res);
  } catch (error) {
    next(error);
  }
};

/**
 * Passes environment variables to pass into the bundle.  They can be accessed via import.meta.env.<name> (or process.env.<name> for compatability)
 */
const environmentVariablesPlugin = env => ({
  name: 'process-global',

  transform(code) {
    for (const [property, value] of Object.entries(env)) {
      code = code.replace(new RegExp(`(?:import\\.meta|process)\\.env\\.${property}`, 'g'), JSON.stringify(value));
    }

    return code;
  }

});

const shouldProcess = id => {
  if (id[0] === '\0') return false;
  return jsExts.test(id);
};

const esbuildPlugin = esbuildOptions => ({
  name: 'esbuild',

  async transform(code, id) {
    if (!shouldProcess(id)) return null;
    const ext = path.extname(id).slice(1);
    const loader = /[cm]?jsx?$/.test(ext) ? 'jsx' : ext;
    return esbuild__namespace.transform(code, {
      sourcefile: id,
      loader,
      sourcemap: 'external',
      ...esbuildOptions
    }).catch(error => {
      if (!('errors' in error)) throw error;
      const err = error.errors[0];
      this.error(err.text, {
        line: err.location.line,
        column: err.location.column
      });
    });
  }

});

let npmCache;
/**
 * Any package names in this set will need to have their named exports detected manually via require()
 * because the export names cannot be statically analyzed
 */

const dynamicCJSModules = new Set(['prop-types', 'react-dom', 'react']);
/**
 * Bundle am npm module entry path into a single file
 * @param mod The full path of the module to bundle, including subpackage/path
 * @param id The imported identifier
 * @param optimize Whether the bundle should be a minified/optimized bundle, or the default quick non-optimized bundle
 */

const bundleNpmModule = async (mod, id, optimize, envVars) => {
  let namedExports = [];

  if (dynamicCJSModules.has(id)) {
    let isValidCJS = true;

    try {
      const text = await node_fs.promises.readFile(mod, 'utf8'); // Goal: Determine if it is ESM or CJS.
      // Try to parse it with cjs-module-lexer, if it fails, assume it is ESM
      // eslint-disable-next-line @cloudfour/typescript-eslint/await-thenable

      await cjsModuleLexer.parse(text);
    } catch {
      isValidCJS = false;
    }

    if (isValidCJS) {
      const require$1 = node_module.createRequire(`file://${__filename}`);

      const imported = require$1(mod);

      if (typeof imported === 'object' && !imported.__esModule) namedExports = Object.keys(imported);
    }
  }

  const virtualEntry = '\0virtualEntry';
  const hasSyntheticNamedExports = namedExports.length > 0;
  const bundle = await rollup.rollup({
    input: hasSyntheticNamedExports ? virtualEntry : mod,
    cache: npmCache,
    shimMissingExports: true,
    treeshake: true,
    preserveEntrySignatures: 'allow-extension',
    plugins: [hasSyntheticNamedExports && {
      // This plugin handles special-case packages whose named exports cannot be found via static analysis
      // For these packages, the package is require()'d, and the named exports are determined that way.
      // A virtual entry exports the named exports from the real entry package
      name: 'cjs-named-exports',

      resolveId(id) {
        if (id === virtualEntry) return virtualEntry;
      },

      load(id) {
        if (id === virtualEntry) {
          const code = `export * from '${mod}'
export {${namedExports.join(', ')}} from '${mod}'
export { default } from '${mod}'`;
          return code;
        }
      }

    }, pluginNodeResolve(), environmentVariablesPlugin(envVars), commonjs__default({
      extensions: ['.js', '.cjs', ''],
      sourceMap: false,
      transformMixedEsModules: true
    }), optimize && {
      name: 'esbuild-minify',
      renderChunk: async code => {
        const output = await esbuild__namespace.transform(code, {
          minify: true,
          legalComments: 'none'
        });
        return {
          code: output.code
        };
      }
    }].filter(Boolean)
  });
  npmCache = bundle.cache;
  const {
    output
  } = await bundle.generate({
    format: 'es',
    indent: false,
    exports: 'named',
    preferConst: true
  });
  return output[0].code;
};

const pluginNodeResolve = () => ({
  name: 'pleasantest-node-resolve',

  resolveId(id) {
    if (isBareImport(id)) return {
      id,
      external: true
    };
  }

});

/**
 * @param {object} exports
 * @param {Set<string>} keys
 */
function loop(exports, keys) {
  if (typeof exports === 'string') {
    return exports;
  }

  if (exports) {
    let idx, tmp;

    if (Array.isArray(exports)) {
      for (idx = 0; idx < exports.length; idx++) {
        if (tmp = loop(exports[idx], keys)) return tmp;
      }
    } else {
      for (idx in exports) {
        if (keys.has(idx)) {
          return loop(exports[idx], keys);
        }
      }
    }
  }
}
/**
 * @param {string} name The package name
 * @param {string} entry The target entry, eg "."
 * @param {number} [condition] Unmatched condition?
 */


function bail(name, entry, condition) {
  throw new Error(condition ? `No known conditions for "${entry}" entry in "${name}" package` : `Missing "${entry}" export in "${name}" package`);
}
/**
 * @param {string} name the package name
 * @param {string} entry the target path/import
 */


function toName(name, entry) {
  return entry === name ? '.' : entry[0] === '.' ? entry : entry.replace(new RegExp('^' + name + '\/'), './');
}
/**
 * @param {object} pkg package.json contents
 * @param {string} [entry] entry name or import path
 * @param {object} [options]
 * @param {boolean} [options.browser]
 * @param {boolean} [options.require]
 * @param {string[]} [options.conditions]
 * @param {boolean} [options.unsafe]
 */


function resolve(pkg, entry = '.', options = {}) {
  let {
    name,
    exports
  } = pkg;

  if (exports) {
    let {
      browser,
      require,
      unsafe,
      conditions = []
    } = options;
    let target = toName(name, entry);
    if (target[0] !== '.') target = './' + target;

    if (typeof exports === 'string') {
      return target === '.' ? exports : bail(name, target);
    }

    let allows = new Set(['default', ...conditions]);
    unsafe || allows.add(require ? 'require' : 'import');
    unsafe || allows.add(browser ? 'browser' : 'node');
    let key,
        tmp,
        isSingle = false;

    for (key in exports) {
      isSingle = key[0] !== '.';
      break;
    }

    if (isSingle) {
      return target === '.' ? loop(exports, allows) || bail(name, target, 1) : bail(name, target);
    }

    if (tmp = exports[target]) {
      return loop(tmp, allows) || bail(name, target, 1);
    }

    for (key in exports) {
      tmp = key[key.length - 1];

      if (tmp === '/' && target.startsWith(key)) {
        return (tmp = loop(exports[key], allows)) ? tmp + target.substring(key.length) : bail(name, target, 1);
      }

      if (tmp === '*' && target.startsWith(key.slice(0, -1))) {
        // do not trigger if no *content* to inject
        if (target.substring(key.length - 1).length > 0) {
          return (tmp = loop(exports[key], allows)) ? tmp.replace('*', target.substring(key.length - 1)) : bail(name, target, 1);
        }
      }
    }

    return bail(name, target);
  }
}
/**
 * @param {object} pkg
 * @param {object} [options]
 * @param {string|boolean} [options.browser]
 * @param {string[]} [options.fields]
 */

function legacy(pkg, options = {}) {
  let i = 0,
      value,
      browser = options.browser,
      fields = options.fields || ['module', 'main'];

  if (browser && !fields.includes('browser')) {
    fields.unshift('browser');
  }

  for (; i < fields.length; i++) {
    if (value = pkg[fields[i]]) {
      if (typeof value == 'string') ; else if (typeof value == 'object' && fields[i] == 'browser') {
        if (typeof browser == 'string') {
          value = value[browser = toName(pkg.name, browser)];
          if (value == null) return browser;
        }
      } else {
        continue;
      }

      return typeof value == 'string' ? './' + value.replace(/^\.?\//, '') : value;
    }
  }
}

const resolveCache = new Map();

const resolveCacheKey = (id, importer, root) => `${id}\n${importer}\n${root}`;

const stat = path => node_fs.promises.stat(path).catch(() => null); // Note: Node does not allow implicit extension resolution for .cjs or .mjs files
// (it also doesn't do implicit extension resolution at all in modules, but it is so common because of bundlers, that we will support it)


const exts = ['.js', '.ts', '.tsx', '.jsx'];

const isFile = async path => {
  const stats = await stat(path);
  return (stats == null ? void 0 : stats.isFile()) || false;
};

const isDirectory = async path => {
  const stats = await stat(path);
  return (stats == null ? void 0 : stats.isDirectory()) || false;
};

const resolveRelativeOrAbsolute = async (id, importer) => {
  // LOAD_AS_FILE section in https://nodejs.org/api/modules.html#modules_all_together
  const resolved = importer ? path.resolve(path.dirname(importer), id) : id;
  const result = await resolveAsFile(resolved);
  if (result) return result; // LOAD_AS_DIRECTORY section in https://nodejs.org/api/modules.html#modules_all_together

  if (await isDirectory(resolved)) {
    const result = await resolveAsDirectory(resolved);
    if (result) return result;
  }

  throw new Error(`Could not resolve ${id}`);
};
/** Given ./file, check for ./file.js (LOAD_AS_FILE) */

const resolveAsFile = async path => {
  if (await isFile(path)) return path;

  for (const ext of exts) {
    if (await isFile(path + ext)) return path + ext;
  }
};
/** Given ./folder, check for ./folder/index.js (LOAD_INDEX) */


const resolveIndex = async directory => {
  const path$1 = path.join(directory, 'index');

  for (const ext of exts) {
    const stats = await stat(path$1 + ext);
    if (stats != null && stats.isFile()) return path$1 + ext;
  }
};

const readPkgJson = async directory => {
  const pkgJsonPath = path.join(directory, 'package.json');

  if (await isFile(pkgJsonPath)) {
    try {
      return JSON.parse(await node_fs.promises.readFile(pkgJsonPath, 'utf8'));
    } catch (error) {
      throw new Error(`Could not read or parse package.json at ${pkgJsonPath}: ${error.message}`);
    }
  }
};

const resolveAsDirectory = async directory => {
  const pkgJson = await readPkgJson(directory);

  if (pkgJson) {
    // Node does not look at the "exports" field when importing a directory via a relative/absolute import.
    // Node only uses the "exports" field when the import is a bare import
    const main = readMainFields(pkgJson, '.', false);

    if (main) {
      const resolvedMain = path.resolve(directory, main);
      const result = (await resolveAsFile(resolvedMain)) || (await resolveIndex(resolvedMain));
      if (result) return result;
    }
  } // Couldn't resolve via ./folder/package.json, so check for ./folder/index.js


  return resolveIndex(directory);
};

const resolveFromNodeModules = async (id, importer, root) => {
  const cacheKey = resolveCacheKey(id, importer, root);
  const cached = resolveCache.get(cacheKey);
  if (cached) return cached;
  const pathChunks = id.split(path.posix.sep);
  const isNpmNamespace = id[0] === '@'; // If it is an npm namespace, then get the first two folders, otherwise just one

  const packageName = pathChunks.slice(0, isNpmNamespace ? 2 : 1); // Path within imported module

  const subPath = path.join(...pathChunks.slice(isNpmNamespace ? 2 : 1));
  const realRoot = await node_fs.promises.realpath(root).catch(() => root); // Walk up folder by folder until a folder is found with <folder>/node_modules/<pkgName>
  // i.e. for 'asdf' from a/b/c.js look at
  // a/b/node_modules/asdf,
  // a/node_modules/asdf,
  // node_modules/asdf,

  let pkgDir;
  let scanDir = importer ? await node_fs.promises.realpath(importer).then(realImporter => path.relative(realRoot, realImporter)).catch(() => path.relative(root, importer)) : '.';

  while (!pkgDir || !(await isDirectory(pkgDir))) {
    if (scanDir === '.' || scanDir.startsWith('..')) {
      throw new Error(`Could not find ${id} in node_modules`);
    } // Not found; go up a level and try again


    scanDir = path.dirname(scanDir);
    pkgDir = path.join(root, scanDir, 'node_modules', ...packageName);
  }

  const pkgJson = await readPkgJson(pkgDir);
  const main = readMainFields(pkgJson, subPath, true);
  let result;
  if (main) result = path.join(pkgDir, main);else if (!('exports' in pkgJson)) {
    const fullPath = path.join(pkgDir, subPath);
    result = (await resolveAsFile(fullPath)) || (await resolveAsDirectory(fullPath));
  }

  if (result) {
    if (!(await stat(result))) throw new Error(`Could not resolve ${id}: ${result} does not exist`);
    const version = pkgJson.version;
    const normalizedPkgName = packageName.join('__');
    const idWithVersion = path.join(version ? `${normalizedPkgName}@${version}` : normalizedPkgName, subPath);
    const resolved = {
      path: await node_fs.promises.realpath(result),
      idWithVersion
    };
    resolveCache.set(cacheKey, resolved);
    return resolved;
  }

  throw new Error(`Could not resolve ${id}: ${pkgDir} exists but no package entrypoint was found`);
};

const readMainFields = (pkgJson, subPath, useExports) => {
  var _resolveLegacy;

  if (useExports && 'exports' in pkgJson) {
    return resolve(pkgJson, subPath, {
      browser: true,
      conditions: ['development', 'esmodules', 'module']
    });
  }

  let result;
  if (subPath === '.') result = legacy(pkgJson, {
    browser: false,
    fields: ['esmodules', 'modern', 'module', 'jsnext:main']
  });
  if (!result) result = (_resolveLegacy = legacy(pkgJson, {
    browser: true,
    fields: []
  })) == null ? void 0 : _resolveLegacy[subPath];
  if (!result && subPath === '.') result = legacy(pkgJson, {
    browser: false,
    fields: ['main']
  });
  return result;
};

// This is the folder that Pleasantest is installed in (e.g. <something>/node_modules/pleasantest)
const installFolder = path.dirname(path.dirname(path.dirname(node_url.fileURLToPath(`file://${__filename}`)))); // Something like <something>/node_modules/pleasantest/.cache

const cacheDir = path.join(installFolder, '.cache');
const npmTranspileCache = new Map();

const setInCache = (cachePath, code) => {
  npmTranspileCache.set(cachePath, code);
  node_fs.promises.mkdir(path.dirname(cachePath), {
    recursive: true
  }).then(() => node_fs.promises.writeFile(cachePath, code));
};

const getFromCache = async cachePath => npmTranspileCache.get(cachePath) || node_fs.promises.readFile(cachePath, 'utf8').catch(() => {}) // Ignore if file is not found in cache
.then(code => {
  if (code) npmTranspileCache.set(cachePath, code);
  return code;
});

const npmPlugin = ({
  root,
  envVars
}) => ({
  name: 'npm',

  // Rewrite bare imports to have @npm/ prefix
  async resolveId(id, importer) {
    if (!isBareImport(id)) return;
    const resolved = await resolveFromNodeModules(id, importer, root).catch(error => {
      throw importer ? changeErrorMessage(error, msg => `${msg} (imported by ${importer})`) : error;
    });
    if (!jsExts.test(resolved.path)) // Don't pre-bundle, use the full path to the file in node_modules
      // (ex: CSS files in node_modules)
      return resolved.path;
    return `${npmPrefix}${id}?resolvedPath=${encodeURIComponent(resolved.path)}&idWithVersion=${encodeURIComponent(resolved.idWithVersion)}`;
  },

  async load(fullId) {
    if (!fullId.startsWith(npmPrefix)) return;
    fullId = fullId.slice(npmPrefix.length);
    const params = new URLSearchParams(fullId.slice(fullId.indexOf('?')));
    const [id] = fullId.split('?');
    let resolvedPath = params.get('resolvedPath');
    let idWithVersion = params.get('idWithVersion');

    if (!resolvedPath || !idWithVersion) {
      const resolveResult = await resolveFromNodeModules(id, undefined, root);
      resolvedPath = resolveResult.path;
      idWithVersion = resolveResult.idWithVersion;
    }

    const cachePath = path.join(cacheDir, '@npm', `${idWithVersion}-${hash(envVars)}.js`);
    const cached = await getFromCache(cachePath);
    if (cached) return cached;
    const result = await bundleNpmModule(resolvedPath, id, false, envVars); // Queue up a second-pass optimized/minified build

    bundleNpmModule(resolvedPath, id, true, envVars).then(optimizedResult => {
      setInCache(cachePath, optimizedResult);
    });
    setInCache(cachePath, result);
    return result;
  }

});

const hash = inputs => node_crypto.createHash('sha512').update(JSON.stringify(inputs)).digest('hex').slice(0, 7);

/**
 * Handles resolving './foo' to './foo.js', and './foo/index.js', and resolving through './foo/package.json'
 */
const resolveExtensionsPlugin = () => ({
  name: 'resolve-extensions-plugin',

  async resolveId(id, importer) {
    if (!isRelativeOrAbsoluteImport(id)) return;
    return resolveRelativeOrAbsolute(id, importer).catch(error => {
      throw importer ? changeErrorMessage(error, msg => `${msg} (imported by ${importer})`) : error;
    });
  }

});

const createModuleServer = async ({
  root = process.cwd(),
  plugins: userPlugins = [],
  envVars: _envVars = {},
  esbuild: esbuildOptions = {}
} = {}) => {
  const prePlugins = [];
  const normalPlugins = [];
  const postPlugins = [];
  const envVars = {
    NODE_ENV: 'development',
    ..._envVars
  };

  for (const plugin of userPlugins) {
    if (!plugin) continue;
    if (plugin.enforce === 'pre') prePlugins.push(plugin);else if (plugin.enforce === 'post') postPlugins.push(plugin);else normalPlugins.push(plugin);
  }

  const plugins = [...prePlugins, ...normalPlugins, resolveExtensionsPlugin(), environmentVariablesPlugin(envVars), npmPlugin({
    root,
    envVars
  }), ...(esbuildOptions ? [esbuildPlugin(esbuildOptions)] : []), cssPlugin({
    root
  }), ...postPlugins];
  const requestCache = new Map();
  const middleware = [indexHTMLMiddleware, await jsMiddleware({
    root,
    plugins,
    requestCache
  }), cssMiddleware({
    root
  }), staticMiddleware({
    root
  })];
  return { ...(await createServer({
      middleware
    })),
    requestCache
  };
};

const queryNames = ['findAllByAltText', 'findAllByDisplayValue', 'findAllByLabelText', 'findAllByPlaceholderText', 'findAllByRole', 'findAllByTestId', 'findAllByText', 'findAllByTitle', 'findByAltText', 'findByDisplayValue', 'findByLabelText', 'findByPlaceholderText', 'findByRole', 'findByTestId', 'findByText', 'findByTitle', 'getAllByAltText', 'getAllByDisplayValue', 'getAllByLabelText', 'getAllByPlaceholderText', 'getAllByRole', 'getAllByTestId', 'getAllByText', 'getAllByTitle', 'getByAltText', 'getByDisplayValue', 'getByLabelText', 'getByPlaceholderText', 'getByRole', 'getByTestId', 'getByText', 'getByTitle', 'queryAllByAltText', 'queryAllByDisplayValue', 'queryAllByLabelText', 'queryAllByPlaceholderText', 'queryAllByRole', 'queryAllByTestId', 'queryAllByText', 'queryAllByTitle', 'queryByAltText', 'queryByDisplayValue', 'queryByLabelText', 'queryByPlaceholderText', 'queryByRole', 'queryByTestId', 'queryByText', 'queryByTitle'];
const getQueriesForElement = (page, asyncHookTracker, element) => {
  const serverPromise = createClientRuntimeServer(); // @ts-expect-error TS doesn't understand the properties coming out of Object.fromEntries

  const queries = Object.fromEntries(queryNames.map(queryName => {
    const query = async (...args) => {
      const serializedArgs = JSON.stringify(args, (_key, value) => {
        if (value instanceof RegExp) {
          return {
            __serialized: 'RegExp',
            source: value.source,
            flags: value.flags
          };
        }

        return value;
      });
      const {
        port
      } = await serverPromise;
      const result = await page.evaluateHandle( // Using new Function to avoid babel transpiling the import
      new Function('argsString', 'element', `return import("http://localhost:${port}/@pleasantest/dom-testing-library")
                .then(async ({ reviveElementsInString, printElement, addToElementCache, ...dtl }) => {
                  const deserializedArgs = JSON.parse(argsString, (key, value) => {
                    if (value.__serialized === 'RegExp')
                      return new RegExp(value.source, value.flags)
                    return value
                  })
                  try {
                    return await dtl.${queryName}(element, ...deserializedArgs)
                  } catch (error) {
                    const message =
                      error.message +
                      (error.container
                        ? '\\n\\nWithin: ' + addToElementCache(error.container)
                        : '')
                    const messageWithElementsRevived = reviveElementsInString(message)
                    const messageWithElementsStringified = messageWithElementsRevived
                      .map(el => {
                        if (el instanceof Element || el instanceof Document)
                          return printElement(el, ${printColorsInErrorMessages})
                        return el
                      })
                      .join('')
                    return { failed: true, messageWithElementsRevived, messageWithElementsStringified }
                  }
                })`), serializedArgs, (element == null ? void 0 : element.asElement()) || (await page.evaluateHandle(() => document)));
      const failed = await result.evaluate(r => typeof r === 'object' && r !== null && r.failed);

      if (failed) {
        const resultProperties = Object.fromEntries(await result.getProperties());
        const messageWithElementsStringified = await resultProperties.messageWithElementsStringified.jsonValue();
        const messageWithElementsRevived = await jsHandleToArray(resultProperties.messageWithElementsRevived);
        const error = new Error(messageWithElementsStringified); // @ts-expect-error messageForBrowser is a custom property that we add to Errors

        error.messageForBrowser = messageWithElementsRevived;
        throw removeFuncFromStackTrace(error, queries[queryName]);
      } // If it returns a JSHandle<Array>, make it into an array of JSHandles so that using [0] for getAllBy* queries works


      if (await result.evaluate(r => Array.isArray(r))) {
        const array = Array.from({
          length: await result.evaluate(r => r.length)
        });
        const props = await result.getProperties();

        for (const [key, value] of props.entries()) {
          array[key] = value;
        }

        return array;
      } // If it is an element, return it


      if (result.asElement() !== null) return result; // Try to JSON-ify it (for example if it is null from queryBy*)

      return result.jsonValue();
    };

    return [queryName, async (...args) => // await is needed for correct stack trace
    // eslint-disable-next-line no-return-await
    await asyncHookTracker.addHook(() => query(...args), queries[queryName])];
  }));
  return queries;
};
let waitForCounter = 0;
const waitFor = async (page, asyncHookTracker, cb, {
  onTimeout,
  container,
  ...opts
}, wrappedFunction) => asyncHookTracker.addHook(async () => {
  const {
    port
  } = await createClientRuntimeServer();
  waitForCounter++; // Functions exposed via page.exposeFunction can't be removed,
  // So we need a unique name for each variable

  const browserFuncName = `pleasantest_waitFor_${waitForCounter}`;
  await page.exposeFunction(browserFuncName, cb);
  const evalResult = await page.evaluateHandle( // Using new Function to avoid babel transpiling the import
  new Function('opts', 'container', `return import("http://localhost:${port}/@pleasantest/dom-testing-library")
          .then(async ({ waitFor }) => {
            try {
              const result = await waitFor(${browserFuncName}, { ...opts, container })
              return { success: true, result }
            } catch (error) {
              if (/timed out in waitFor/i.test(error.message)) {
                // Leave out stack trace so the stack trace is given from Node
                return { success: false, result: { message: error.message } }
              }
              return { success: false, result: error }
            }
          })`), opts, // Container has to be passed separately because puppeteer won't unwrap nested JSHandles
  container);
  const wasSuccessful = await evalResult.evaluate(r => r.success);
  const result = await evalResult.evaluate(r => r.success ? r.result : {
    message: r.result.message,
    stack: r.result.stack
  });
  if (wasSuccessful) return result;
  const err = new Error(result.message);
  if (result.stack) err.stack = result.stack;else removeFuncFromStackTrace(err, asyncHookTracker.addHook);
  throw onTimeout ? onTimeout(err) : err;
}, wrappedFunction);

const sourceMapErrorFromBrowser = async (res, requestCache, port, func) => {
  if (res === undefined) return;
  if (typeof res !== 'object' || !res || !('message' in res) || !('stack' in res)) throw res;
  const {
    message,
    stack
  } = res;
  const parsedStack = parseStackTrace(stack);
  const modifiedStack = await Promise.all(parsedStack.map(async stackItem => {
    var _sourceLocation$line;

    if (stackItem.raw.startsWith(stack.slice(0, stack.indexOf('\n')))) return null;
    if (!stackItem.fileName) return stackItem.raw;
    const fileName = stackItem.fileName;
    const line = stackItem.line;
    const column = stackItem.column;
    if (!fileName.startsWith(`http://localhost:${port}`)) return stackItem.raw;
    const url = new URL(fileName);
    const osPath = url.pathname.slice(1).split(path.posix.sep).join(path.sep); // Absolute file path

    const file = path.resolve(process.cwd(), osPath); // Rollup-style Unix-normalized path "id":

    const id = file.split(path.sep).join(path.posix.sep);
    const transformResult = requestCache.get(id);
    const map = typeof transformResult === 'object' && transformResult.map;

    if (!map) {
      let p = url.pathname;
      const npmPrefix = '/@npm/';
      if (p.startsWith(npmPrefix)) p = path.join(process.cwd(), 'node_modules', p.slice(npmPrefix.length));
      return printStackLine(p, line, column, stackItem.name);
    }

    const {
      SourceMapConsumer
    } = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require('source-map')); });
    const consumer = await new SourceMapConsumer(map);
    const sourceLocation = consumer.originalPositionFor({
      line,
      column: column - 1 // Source-map uses zero-based column numbers

    });
    consumer.destroy();
    return printStackLine(path.join(process.cwd(), url.pathname), (_sourceLocation$line = sourceLocation.line) != null ? _sourceLocation$line : line, sourceLocation.column === null ? column : // Convert back from zero-based column to 1-based
    sourceLocation.column + 1, stackItem.name);
  }));
  const errorName = stack.slice(0, stack.indexOf(':')) || 'Error';
  const ErrorConstructor = specializedErrors[errorName] || Error;
  const error = new ErrorConstructor(message);
  const finalStack = modifiedStack.filter(Boolean).join('\n'); // If the browser error did not provide a stack, use the stack trace from node

  if (finalStack) {
    error.stack = `${errorName}: ${message}\n${finalStack}`;
  } else {
    removeFuncFromStackTrace(error, func);
    if (error.stack) error.stack = error.stack.split('\n').filter( // This was appearing in stack traces and it messed up the Jest output
    line => !(/runMicrotasks/.test(line) && /<anonymous>/.test(line))).join('\n');
  }

  throw error;
};
const specializedErrors = {
  EvalError,
  RangeError,
  ReferenceError,
  SyntaxError,
  TypeError,
  URIError
};

/** Wraps each user method to catch errors that happen when user forgets to await */
const wrapWithForgotAwait = (user, asyncHookTracker) => {
  for (const key of Object.keys(user)) {
    const original = user[key]; // eslint-disable-next-line @cloudfour/unicorn/consistent-function-scoping

    const wrapper = async (...args) => // The await is necessary so `wrapper` is in the stack trace,
    // so captureStackTrace works correctly
    // eslint-disable-next-line no-return-await
    await asyncHookTracker.addHook(() => original(...args), wrapper);

    user[key] = wrapper;
  }
};

const pleasantestUser = async (page, asyncHookTracker, userOpts) => {
  const {
    port
  } = await createClientRuntimeServer();

  const runWithUtils = fn => new Function('...args', `return import("http://localhost:${port}/@pleasantest/user-util")
      .then((utils) => {
        try {
          return [utils, (0, ${fn.toString()})(utils, ...args)]
        } catch (error) {
          if (error.error) error = error.error
          return [utils, { error }]
        }
      })
      .then(([utils, result]) => {
        if (result && typeof result === 'object' && result.error) {
          const msgWithLiveEls = result.error
          if (typeof msgWithLiveEls === 'string') return { error: msgWithLiveEls }
          const msgWithStringEls = msgWithLiveEls
            .map(el => {
              if (el instanceof Element || el instanceof Document)
                return utils.printElement(el, ${printColorsInErrorMessages})
              return el
            })
            .join('')
          return { error: { msgWithLiveEls, msgWithStringEls } }
        }
        return result
      })`);

  const user = {
    async click(el, {
      force = false,
      targetSize = userOpts.targetSize
    } = {}) {
      assertElementHandle(el, user.click);
      await el.evaluateHandle(runWithUtils((utils, clickEl, force, targetSize) => {
        utils.assertAttached(clickEl);

        if (!force) {
          if (targetSize !== false) {
            utils.assertTargetSize(clickEl, targetSize);
          }

          utils.assertVisible(clickEl);
          const clickElRect = clickEl.getBoundingClientRect(); // See if there is an element covering the center of the click target element

          const coveringEl = document.elementFromPoint(Math.floor(clickElRect.x + clickElRect.width / 2), Math.floor(clickElRect.y + clickElRect.height / 2));
          if (coveringEl === clickEl || clickEl.contains(coveringEl)) return; // TODO: try to find other points on the element that are clickable,
          // in case the covering element does not cover the whole click-target element

          return utils.error`Could not click element:
${clickEl}

Element was covered by:
${coveringEl}`;
        }
      }), force, targetSize).then(throwBrowserError(user.click));
      await el.click();
    },

    // Implementation notes:
    // - *Appends* to existing text - different from some other tools:
    //   Puppeteer and Playwright *Prepend*
    //   Cypress and user-event *Append*
    // - The names of the commands in curly brackets are mirroring the user-event command names
    //   *NOT* the Cypress names.
    //   i.e. Cypress uses {leftarrow} but user-event and pleasantest use {arrowleft}
    async type(el, text, {
      delay = 1,
      force = false
    } = {}) {
      assertElementHandle(el, user.type); // Splits input into chunks
      // i.e. "something{backspace}something{enter} "
      // => ["something", "{backspace}", "something", "{enter}"]

      const chunks = [];

      while (text) {
        text = text.startsWith('{') ? text.replace(/^{[^{}]*}/, substr => {
          chunks.push(substr.toLowerCase());
          return '';
        }) : text.replace(/^[^{]*/, substr => {
          chunks.push(substr);
          return '';
        });
      }

      await el.evaluateHandle(runWithUtils((utils, el, force) => {
        utils.assertAttached(el);
        if (!force) utils.assertVisible(el);

        if (el instanceof HTMLInputElement || el instanceof HTMLTextAreaElement) {
          // No need to focus it if it is already focused
          // We won't move the cursor to the end either because that could be unexpected
          if (document.activeElement === el) return;
          el.focus(); // Move cursor to the end

          const end = el.value.length;
          el.setSelectionRange(end, end);
        } else if (el instanceof HTMLElement && el.isContentEditable) {
          // No need to focus it if it is already focused
          // We won't move the cursor to the end either because that could be unexpected
          if (document.activeElement === el) return;
          el.focus();
          const range = el.ownerDocument.createRange();
          range.selectNodeContents(el);
          const end = el.textContent.length;
          range.setStart(el.firstChild, end);
          range.setEnd(el.firstChild, end); // Move cursor to the end

          const sel = el.ownerDocument.getSelection();
          sel.removeAllRanges();
          sel.addRange(range);
        } else {
          return utils.error`Cannot type in element that is not typeable:
${el}
Element must be an <input> or <textarea> or an element with the contenteditable attribute.`;
        }
      }), force).then(throwBrowserError(user.type));

      for (const chunk of chunks) {
        const key = typeCommandsMap[chunk];

        if (key) {
          await page.keyboard.press(key, {
            delay
          });
        } else if (chunk === '{selectall}') {
          await el.evaluateHandle(runWithUtils((utils, el) => {
            if (el instanceof HTMLInputElement || el instanceof HTMLTextAreaElement) {
              el.select();
            } else {
              return utils.error`{selectall} command is only available for <input> and <textarea> elements, received: ${el}`;
            }
          })).then(throwBrowserError(user.type));
        } else {
          await page.keyboard.type(chunk, {
            delay
          });
        }
      }
    },

    async clear(el, {
      force = false
    } = {}) {
      assertElementHandle(el, () => user.clear);
      await el.evaluateHandle(runWithUtils((utils, el, force) => {
        utils.assertAttached(el);
        if (!force) utils.assertVisible(el);

        if (el instanceof HTMLInputElement || el instanceof HTMLTextAreaElement) {
          el.select();
        } else {
          return utils.error`user.clear is only available for <input> and <textarea> elements, received: ${el}`;
        }
      }), force).then(throwBrowserError(user.clear));
      await page.keyboard.press('Backspace');
    },

    async selectOptions(el, values, {
      force = false
    } = {}) {
      assertElementHandle(el, user.selectOptions);
      const valuesArray = Array.isArray(values) ? values : [values];

      for (const value of valuesArray) {
        // Make sure all values are strings or ElementHandles
        if (typeof value !== 'string') {
          assertElementHandle(value, user.selectOptions, 'values must be a string or ElementHandle or array of either of those.');
        }
      }

      const valuesArrayHandle = await el.evaluateHandle(runWithUtils((utils, el, force, ...valuesArray) => {
        utils.assertAttached(el);
        if (!force) utils.assertVisible(el);
        if (!(el instanceof HTMLSelectElement)) return utils.error`user.selectOptions is only available for <select> elements, received: ${el}`;
        if (valuesArray.length > 1 && !el.multiple) return utils.error`Cannot select multiple options on a <select> element without the \`multiple\` attribute:\n\n${el}`;
        const validOptions = new Set([...el.options].map(el => el.value));
        return valuesArray.map(value => {
          if (value instanceof HTMLOptionElement) {
            if (!validOptions.has(value.value) || ![...el.options].includes(value)) {
              throw utils.error`Could not select an option ${value}, it is not one of the valid options in the <select>. Valid options are: ${JSON.stringify([...validOptions])}`;
            }

            return value.value;
          }

          if (!validOptions.has(value)) throw utils.error`Could not select an option ${JSON.stringify(value)}, it is not one of the valid options in the <select>. Valid options are: ${JSON.stringify([...validOptions])}`;
          return value;
        });
      }), force, ...valuesArray).then(throwBrowserError(user.selectOptions));
      await el.select(...(await valuesArrayHandle.jsonValue()));
    }

  };
  wrapWithForgotAwait(user, asyncHookTracker);
  return user;
}; // Note: command chunks are already lowercased so it is not case-sensitive
// left side: command
// right side: passed to pptr keyboard.press, full list at https://github.com/puppeteer/puppeteer/blob/main/src/common/USKeyboardLayout.ts

const typeCommandsMap = {
  '{enter}': 'Enter',
  '{arrowleft}': 'ArrowLeft',
  '{arrowright}': 'ArrowRight',
  '{arrowup}': 'ArrowLeft',
  '{arrowdown}': 'ArrowDown',
  '{tab}': 'Tab',
  '{backspace}': 'Backspace',
  '{del}': 'Delete'
};
/**
 * When code that is evaluated in the browser returns {error: ...}, this function causes an error to be thrown in Node.
 * This is better than just throwing directly from the browser code because that would cause the error to be wrapped by Puppeteer's EvaluationError, which causes a confusing stack trace.
 */

const throwBrowserError = func => async result => {
  const resultJSON = await result.jsonValue();

  if (resultJSON != null && resultJSON.error) {
    let err;

    if (typeof resultJSON.error === 'string') {
      err = new Error(resultJSON.error);
    } else {
      const errorProp = await result.getProperty('error');
      const errorProperties = Object.fromEntries(await errorProp.getProperties());

      if ( // eslint-disable-next-line @cloudfour/typescript-eslint/no-unnecessary-condition
      errorProperties.msgWithStringEls && // eslint-disable-next-line @cloudfour/typescript-eslint/no-unnecessary-condition
      errorProperties.msgWithLiveEls) {
        err = new Error(await errorProperties.msgWithStringEls.jsonValue()); // @ts-expect-error messageForBrowser is a custom thing

        err.messageForBrowser = await jsHandleToArray(errorProperties.msgWithLiveEls);
      } else {
        err = new Error(await errorProp.jsonValue());
      }
    }

    removeFuncFromStackTrace(err, func);
    throw err;
  }

  return result;
};

const regexHandler = {
  name: 'RegExp',
  detect: input => input instanceof RegExp,
  toObj: input => ({
    source: input.source,
    flags: input.flags
  }),
  fromObj: input => new RegExp(input.source, input.flags)
};
const handlers = [regexHandler];
/**
 * @param input The data to serialize
 * @param visitor Function gets called on every value before serialization. Return the value unchanged, or modify and return it.
 */

const serialize = (input, visitor) => JSON.stringify(input, (_key, _value) => {
  const value = visitor ? visitor(_value) : _value;
  const handler = handlers.find(h => h.detect(value));
  if (handler) return {
    __objType: handler.name,
    value: handler.toObj(value)
  };
  return value;
});
const deserialize = input => JSON.parse(input, (_key, value) => {
  var _handlers$find;

  const serializedType = typeof value === 'object' && (value == null ? void 0 : value.__objType);
  return serializedType && ((_handlers$find = handlers.find(h => h.name === serializedType)) == null ? void 0 : _handlers$find.fromObj(value.value)) || value;
});

const methods = ['toBeInTheDocument', 'toBeEmptyDOMElement', 'toContainElement', 'toContainHTML', 'toHaveAccessibleDescription', 'toHaveAccessibleName', 'toHaveTextContent', 'toHaveAttribute', 'toHaveClass', 'toHaveStyle', 'toHaveFocus', 'toHaveFormValues', 'toBeVisible', 'toBeDisabled', 'toBeEnabled', 'toBeRequired', 'toBeInvalid', 'toBeValid', 'toHaveValue', 'toHaveDisplayValue', 'toBeChecked', 'toBePartiallyChecked', 'toHaveErrorMessage'];

const isJSHandle = input => {
  if (typeof input !== 'object' || !input) return false; // @ts-expect-error checking for properties that don't necessarily exist

  return input.asElement && input.dispose && input.evaluate;
};

const matchers = Object.fromEntries(methods.map(methodName => {
  const matcher = async function (elementHandle, ...matcherArgs) {
    const serverPromise = createClientRuntimeServer();

    if (!isElementHandle(elementHandle)) {
      // Special case: expect(null).not.toBeInTheDocument() should pass
      if (methodName === 'toBeInTheDocument' && this.isNot) {
        // This is actually passing but since it is isNot it has to return false
        return {
          pass: false,
          message: () => ''
        };
      }

      const message = [this.utils.matcherHint(`${this.isNot ? '.not' : ''}.${methodName}`, 'received', ''), '', `${this.utils.RECEIVED_COLOR('received')} value must be an HTMLElement or an SVGElement.`, isPromise(elementHandle) ? `Received a ${this.utils.RECEIVED_COLOR('Promise')}. Did you forget to await?` : this.utils.printWithType('Received', elementHandle, this.utils.printReceived)].join('\n');
      throw removeFuncFromStackTrace(new Error(message), matcher);
    }

    for (const arg of matcherArgs) {
      if (typeof arg === 'object' && typeof (arg == null ? void 0 : arg.asymmetricMatch) === 'function') {
        const error = new Error(`Pleasantest does not support using asymmetric matchers in browser-based matchers

Received ${this.utils.printReceived(arg)}`);
        throw removeFuncFromStackTrace(error, matcher);
      }
    }

    const {
      port
    } = await serverPromise;
    const ctxString = JSON.stringify(this); // Contains stuff like isNot and promise

    const result = await elementHandle.evaluateHandle( // Using new Function to avoid babel transpiling the import
    new Function('element', '...matcherArgs', `return import("http://localhost:${port}/@pleasantest/jest-dom")
            .then(({ jestContext, deserialize, ...jestDom }) => {
            const context = { ...(${ctxString}), ...jestContext }
            try {
              const deserialized = matcherArgs
              .slice(1)
              .map(a => typeof a === 'string' ? deserialize(a) : a)
              return jestDom.${methodName}.call(context, element, ...deserialized)
            } catch (error) {
              return { thrown: true, error }
            }
          })`), elementHandle, ...matcherArgs.map(arg => isJSHandle(arg) ? arg : serialize(arg))); // Whether the matcher threw (this is different from the matcher failing)
    // The matcher failing means that it returned a result for Jest to throw
    // But a matcher throwing means that the input was invalid or something

    const thrownError = await result.evaluate(result => result.thrown); // We have to evaluate the message right away
    // because Jest does not accept a promise from the returned message property

    const message = await result.evaluate(thrownError ? matcherResult => matcherResult.error.message : matcherResult => matcherResult.message());
    const deserializedMessage = runJestUtilsInNode(message, this);
    const {
      messageWithElementsRevived,
      messageWithElementsStringified
    } = await elementHandle.evaluateHandle(new Function('el', 'message', `return import("http://localhost:${port}/@pleasantest/jest-dom")
                .then(({ reviveElementsInString, printElement }) => {
                  const messageWithElementsRevived = reviveElementsInString(message)
                  const messageWithElementsStringified = messageWithElementsRevived
                  .map(el => {
                    if (el instanceof Element) return printElement(el, ${printColorsInErrorMessages})
                      return el
                  })
                  .join('')
                  return { messageWithElementsRevived, messageWithElementsStringified }
                })`), deserializedMessage).then(async returnHandle => {
      const {
        messageWithElementsRevived,
        messageWithElementsStringified
      } = Object.fromEntries(await returnHandle.getProperties());
      return {
        messageWithElementsStringified: await messageWithElementsStringified.jsonValue(),
        messageWithElementsRevived: await jsHandleToArray(messageWithElementsRevived)
      };
    });

    if (thrownError) {
      const error = new Error(messageWithElementsStringified); // @ts-expect-error messageForBrowser is a property we added to Error

      error.messageForBrowser = messageWithElementsRevived;
      throw removeFuncFromStackTrace(error, matcher);
    }

    return { ...(await result.jsonValue()),
      message: () => messageWithElementsStringified,
      messageForBrowser: messageWithElementsRevived
    };
  };

  const matcherWrapper = async function (elementHandle, ...matcherArgs) {
    const asyncHookTracker = activeAsyncHookTrackers.size === 1 && activeAsyncHookTrackers[Symbol.iterator]().next().value;

    if (asyncHookTracker) {
      const res = await asyncHookTracker.addHook(() => matcher.call(this, elementHandle, ...matcherArgs), matchers[methodName]); // AddHook resolves to undefined if the function throws after the async hook tracker closes
      // Because it needs to not trigger an unhandled promise rejection
      // asyncHookTracker can return undefined (even though its types say it won't)
      // if the user forgot to use await,
      // and the test already exited/threw because of the withBrowser forgot-await detection,
      // but the code will keep running because it's impossible to stop without an unhandled promise rejection,
      // which is frustrating to debug
      // eslint-disable-next-line @cloudfour/typescript-eslint/no-unnecessary-condition

      if (res === undefined) return {
        pass: !this.isNot,
        message: () => ''
      };
      return res;
    }

    return matcher.call(this, elementHandle, ...matcherArgs);
  };

  return [methodName, matcherWrapper];
}));

const runJestUtilsInNode = (message, context) => {
  // Handling nested JEST_UTILS calls here is the complexity
  // The while loop goes through them in reverse
  // so inner (nested) calls are evaluated before outer calls
  const jestUtilsCalls = [...message.matchAll(/\$JEST_UTILS\.([$A-Z_a-z]*)\$/g)];
  const closeRegex = /\$END_JEST_UTILS\$/g;
  let jestUtilsCall;

  while (jestUtilsCall = jestUtilsCalls.pop()) {
    var _closeRegex$exec;

    const start = jestUtilsCall.index;
    const methodName = jestUtilsCall[1];
    closeRegex.lastIndex = start;
    const closeIndex = (_closeRegex$exec = closeRegex.exec(message)) == null ? void 0 : _closeRegex$exec.index;

    if (closeIndex !== undefined) {
      const argsString = message.slice(start + jestUtilsCall[0].length, closeIndex);
      const parsedArgs = deserialize(argsString); // @ts-expect-error TS doesn't know about the properties

      const res = context.utils[methodName](...parsedArgs); // Const escaped = res.replace(/"/g, '\\"').replace(/\u001b/g, '\\u001b');

      const escaped = JSON.stringify(res).replace(/^"/, '').replace(/"$/, '');
      message = message.slice(0, start) + escaped + message.slice(closeIndex + '$END_JEST_UTILS$'.length);
    }
  }

  return message.replace(/\\u[\dA-Fa-f]{4}/g, match => JSON.parse(`"${match}"`)).replace(/\\./g, match => JSON.parse(`"${match}"`));
};

expect.extend(matchers); // These type definitions are incomplete, only including methods we've tested
// More can be added from https://unpkg.com/@types/testing-library__jest-dom/index.d.ts
// You can copy-paste and change the return types to promises

const accessibilityTreeSymbol = Symbol('PT Accessibility Tree');

const getAccessibilityTree = async (element, options = {}) => {
  if ( // eslint-disable-next-line @cloudfour/typescript-eslint/no-unnecessary-condition
  element && typeof element === 'object' && element.constructor.name === 'Page') {
    element = await element.evaluateHandle(() => document.documentElement);
  }

  const serverPromise = createClientRuntimeServer();
  assertElementHandle(element, getAccessibilityTreeWrapper);
  const {
    port
  } = await serverPromise;
  const result = await element.evaluate( // Using new Function to avoid babel transpiling the import
  new Function('element', 'options', `return import("http://localhost:${port}/@pleasantest/accessibility")
        .then(accessibility => accessibility.getAccessibilityTree(element, options))`), options);
  return {
    [accessibilityTreeSymbol]: result,
    toString: () => result
  };
}; // Wrapped version adds forgot await checks


const getAccessibilityTreeWrapper = async (...args) => {
  const asyncHookTracker = activeAsyncHookTrackers.size === 1 && activeAsyncHookTrackers[Symbol.iterator]().next().value;

  if (asyncHookTracker) {
    return asyncHookTracker.addHook(() => getAccessibilityTree(...args), getAccessibilityTreeWrapper);
  }

  return getAccessibilityTree(...args);
};
const accessibilityTreeSnapshotSerializer = {
  serialize: (val, config, indentation, depth, refs, printer) => {
    const v = val[accessibilityTreeSymbol];
    return typeof v === 'string' ? v : printer(v, config, indentation, depth, refs);
  },
  test: val => val && typeof val === 'object' && accessibilityTreeSymbol in val
}; // This tells Jest how to print the accessibility tree (without adding extra quotes)
// https://jestjs.io/docs/expect#expectaddsnapshotserializerserializer

expect.addSnapshotSerializer(accessibilityTreeSnapshotSerializer); // Based on https://github.com/WordPress/gutenberg/blob/3b2eccc289cfc90bd99252b12fc4c6e470ce4c04/packages/jest-puppeteer-axe/src/index.js

/** Formats the list of violations object returned by Axe analysis. */

async function formatViolations(violations, page) {
  const {
    port
  } = await createClientRuntimeServer();
  const formattedHandle = await page.evaluateHandle(violations => {
    const output = []; // eslint-disable-next-line @cloudfour/unicorn/consistent-function-scoping

    const findElement = node => [...document.querySelectorAll(node.target)].find(el => el.outerHTML === node.html || // For long strings, axe-core returns only the opening tag
    // https://github.com/dequelabs/axe-core/blob/v4.4.2/lib/core/utils/dq-element.js#L11
    el.outerHTML.slice(0, el.outerHTML.indexOf('>') + 1) === node.html);

    for (const {
      help,
      helpUrl,
      id,
      nodes
    } of violations) {
      // The dollar signs are used to indicate which part should be bolded/red
      // We can't directly call the color functions here since we haven't imported them
      // And this function is toString'd into the browser.
      output.push(`
$$$${help}$$$ (${id})
${helpUrl}
Affected Nodes:
`);

      for (const node of nodes) {
        output.push('\n', findElement(node) || node.html, '\n');

        if (node.any.length > 0) {
          output.push(`Fix ${node.any.length > 1 ? 'any of the following' : 'the following'}:\n`);

          for (const item of node.any) output.push(`  • ${item.message}\n`);
        }

        if (node.all.length > 0 || node.none.length > 0) {
          output.push(`Fix ${node.all.length > 1 ? 'all of the following' : 'the following'}:\n`);

          for (const item of [...node.all, ...node.none]) output.push(`  • ${item.message}.\n`);
        }
      }
    }

    return output;
  }, violations);
  const outputHandle = await page.evaluateHandle( // Using new Function to avoid babel transpiling the import
  new Function('formattedArr', `return import("http://localhost:${port}/@pleasantest/accessibility")
        .then(({ printElement, colors }) => {
          let messageWithElementsStringified = '';
          const messageWithElementsRevived = [];
          for (let chunk of formattedArr) {
            if (typeof chunk === 'string') {
              chunk = chunk.replace(
                /\\$\\$\\$(.*?)\\$\\$\\$/g,
                (_, ruleText) => colors.red(colors.bold(ruleText))
              )
            }
            messageWithElementsRevived.push(chunk);
            if (typeof chunk === 'string') {
              messageWithElementsStringified += chunk;
            } else {
              messageWithElementsStringified += printElement(chunk, ${printColorsInErrorMessages});
            }
          }

          return { messageWithElementsStringified, messageWithElementsRevived };
        })`), formattedHandle);
  const {
    messageWithElementsRevived,
    messageWithElementsStringified
  } = Object.fromEntries(await outputHandle.getProperties());
  return {
    messageWithElementsStringified: await messageWithElementsStringified.jsonValue(),
    messageWithElementsRevived: await jsHandleToArray(messageWithElementsRevived)
  };
}

/**
 * Defines async matcher to check whether a given Puppeteer's page instance passes Axe accessibility tests.
 */
async function toPassAxeTests(page, {
  include,
  exclude,
  disabledRules,
  options,
  config
} = {}) {
  let AxePuppeteer;

  try {
    const axePuppeteerModule = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require('@axe-core/puppeteer')); });
    AxePuppeteer = axePuppeteerModule.AxePuppeteer;
  } catch {
    throw removeFuncFromStackTrace(new Error('Install @axe-core/puppeteer and axe-core to use the toPassAxeTests matcher'), toPassAxeTests);
  }

  const axe = new AxePuppeteer(page);
  if (include) axe.include(include);
  if (exclude) axe.exclude(exclude);
  if (options) axe.options(options);
  if (disabledRules) axe.disableRules(disabledRules);
  if (config) axe.configure(config);
  const {
    violations
  } = await axe.analyze();
  const formattedViolations = await formatViolations(violations, page);
  const pass = violations.length === 0;
  const foundViolationsMessage = `${this.utils.matcherHint('.toPassAxeTests', 'page', '')}
Expected page to pass Axe accessibility tests.
Violations found:
`;
  const expectedViolationsMessage = `${this.utils.matcherHint('.not.toPassAxeTests', 'page', '')}

Expected page to contain accessibility check violations.
No violations found.`;
  const message = pass ? () => expectedViolationsMessage : () => foundViolationsMessage + formattedViolations.messageWithElementsStringified;
  const output = {
    message,
    pass
  };

  if (!pass) {
    // @ts-expect-error This is a custom property we are using to customize the message inside the browser
    output.messageForBrowser = [foundViolationsMessage, ...formattedViolations.messageWithElementsRevived];
  }

  return output;
}

expect.extend({
  toPassAxeTests
});

options.enabled = true;

const ansiRegex = ansiRegex$1({
  onlyFirst: true
});

// Call signatures of withBrowser:
// withBrowser(() => {})
// withBrowser({ ... }, () => {})
// withBrowser.headed(() => {})
// withBrowser.headed({ ... }, () => {})
const withBrowser = (...args) => {
  const testFn = args.length === 1 ? args[0] : args[1];
  const options = args.length === 1 ? {} : args[0];
  const thisFile = node_url.fileURLToPath(`file://${__filename}`); // Figure out the file that called withBrowser so that we can resolve paths correctly from there
  // eslint-disable-next-line @cloudfour/unicorn/error-message

  const stack = parseStackTrace(new Error().stack).map(stackFrame => {
    var _$exec;

    if (stackFrame.fileName) return stackFrame.fileName;
    return (_$exec = /\s*at\s+([\w./-]*)/.exec(stackFrame.raw)) == null ? void 0 : _$exec[1];
  });
  const testFile = stack.find(stackItem => {
    if (!stackItem) return false; // ignore if it is the current file

    if (stackItem === thisFile) return false; // ignore if it is an internal-to-node thing

    if (!stackItem.startsWith('/')) return false; // Find the first item that is not the current file

    return true;
  });
  const testPath = testFile ? path.relative(process.cwd(), testFile) : thisFile;
  return async () => {
    const {
      cleanupServer,
      asyncHookTracker,
      ...ctx
    } = await createTab({
      testPath,
      options
    });

    const cleanup = async leaveOpen => {
      if (!leaveOpen || options.headless) {
        await ctx.page.close();
      }

      ctx.page.browser().disconnect();
      await cleanupServer();
    };

    try {
      await testFn(ctx);
      const forgotAwaitError = asyncHookTracker.close();
      if (forgotAwaitError) throw forgotAwaitError;
    } catch (error) {
      var _error$matcherResult;

      const forgotAwaitError = asyncHookTracker.close();

      if (forgotAwaitError) {
        await cleanup(false);
        throw forgotAwaitError;
      }

      const messageForBrowser = // This is how we attach the elements to the error from testing-library
      (error == null ? void 0 : error.messageForBrowser) || ( // This is how we attach the elements to the error from jest-dom
      error == null ? void 0 : (_error$matcherResult = error.matcherResult) == null ? void 0 : _error$matcherResult.messageForBrowser); // Jest hangs when sending the error
      // from the worker process up to the main process
      // if the error has circular references in it
      // (which it does if there are elementHandles)

      if (error.matcherResult) delete error.matcherResult.messageForBrowser;
      delete error.messageForBrowser;

      if (!options.headless) {
        const failureMessage = [`${bold(white(bgRed(' FAIL ')))}\n\n`];
        const testName = getTestName();

        if (testName) {
          failureMessage.push(`${bold(red(`● ${testName}`))}\n\n`);
        }

        if (messageForBrowser) {
          // Combine adjacent string chunks
          // So the browser does not insert extra line breaks/indents between the lines
          const normalizedMessageForBrowser = messageForBrowser.reduce((acc, chunk) => {
            if (typeof chunk !== 'object' && acc.length > 0 && typeof acc[acc.length - 1] === 'string') acc[acc.length - 1] += String(chunk);else acc.push(typeof chunk === 'object' ? chunk : String(chunk));
            return acc;
          }, []);
          failureMessage.push(...normalizedMessageForBrowser.map((segment, i) => {
            if (typeof segment !== 'string') return segment;

            if (i !== 0 && typeof normalizedMessageForBrowser[i - 1] !== 'string') {
              return indent(segment, false);
            }

            return indent(segment);
          }));
        } else {
          failureMessage.push(indent(error instanceof Error ? error.message : String(error)));
        }

        await ctx.page.evaluate((...colorErr) => {
          console.log(...colorErr);
        }, ...ansiColorsLog(...failureMessage));
      }

      await cleanup(true);
      throw error;
    }

    await cleanup(false);
  };
};

withBrowser.headed = (...args) => {
  const testFn = args.length === 1 ? args[0] : args[1];
  const options = args.length === 1 ? {} : args[0];
  return withBrowser({ ...options,
    headless: false
  }, testFn);
};

const getTestName = () => {
  try {
    return expect.getState().currentTestName;
  } catch {
    return null;
  }
};

const indent = (input, indentFirstLine = true) => input.split('\n').map((line, i) => {
  if (!indentFirstLine && i === 0) return line; // If there is an escape code at the beginning of the line
  // put the tab after the escape code
  // the reason for this is to prevent the indentation from getting messed up from wrapping
  // you can see this if you squish the devtools window

  const match = ansiRegex.exec(line);
  if (!match || match.index !== 0) return `  ${line}`;
  const insertPoint = match[0].length;
  return `${line.slice(0, insertPoint)}  ${line.slice(insertPoint)}`;
}).join('\n');

const createTab = async ({
  testPath,
  options: {
    headless = (() => {
      var _defaultOptions$headl;

      return (_defaultOptions$headl = defaultOptions.headless) != null ? _defaultOptions$headl : true;
    })(),
    device = defaultOptions.device,
    moduleServer: moduleServerOpts = {},
    user: userOpts = {}
  }
}) => {
  const asyncHookTracker = createAsyncHookTracker();
  const browser = await connectToBrowser('chromium', headless);
  const browserContext = await browser.createIncognitoBrowserContext();
  const page = await browserContext.newPage();
  const {
    requestCache,
    port,
    close: closeServer
  } = await createModuleServer({ ...defaultOptions.moduleServer,
    ...moduleServerOpts,
    ...defaultOptions.moduleServer,
    ...moduleServerOpts
  });

  if (device) {
    if (!headless) {
      const session = await page.target().createCDPSession();
      const {
        windowId
      } = await session.send('Browser.getWindowForTarget');
      await session.send('Browser.setWindowBounds', {
        windowId,
        bounds: {
          // Allow space for devtools
          // start-disowned-browser.ts sets the devtools preferences with default width
          width: device.viewport.width + 450,
          height: device.viewport.height + 79 // Allow space for toolbar

        }
      });
      await session.detach();
    }

    await page.emulate(device);
  }

  page.on('console', message => {
    var _message$location$url;

    const text = message.text(); // This is naive, there is probably something better to check
    // If the text includes %c, then it probably came from the jest output being forwarded into the browser
    // So we don't need to print it _again_ in node, since it already came from node

    if (text.includes('%c')) return; // This is intended so that transpilation errors from the module server,
    // which will get a nice code frame in node,
    // do not also log "Failed to load resource: the server responded with a status of 500"

    if (/Failed to load resource: the server responded with a status of 500/.test(text) && (_message$location$url = message.location().url) != null && _message$location$url.includes(`http://localhost:${port}`)) return;
    const type = message.type(); // Create a new console instance instead of using the global one
    // Because the global one is overridden by Jest, and it adds a misleading second stack trace and code frame below it

    const console = new node_console.Console(process.stdout, process.stderr);

    if (type === 'error') {
      const error = new Error(text);
      const location = message.location();
      error.stack = `Error: ${text}
    at ${location.url}`;
      console.error('[browser]', error);
    } else {
      console.log('[browser]', text);
    }
  });
  await page.goto(`http://localhost:${port}`); // eslint-disable-next-line @cloudfour/typescript-eslint/ban-types

  const functionArgs = [];

  const runJS = (code, args) => asyncHookTracker.addHook(async () => {
    await page.exposeFunction('_pleasantestCallFunction', (id, args) => functionArgs[id](...args)).catch(error => {
      if (!error.message.includes('already exists')) throw error;
    }); // For some reason encodeURIComponent doesn't encode '

    const encodedCode = encodeURIComponent(code).replace(/'/g, '%27');
    const buildStatus = createBuildStatusTracker();
    const argsWithFuncsAsObjs = args == null ? void 0 : args.map(arg => {
      if (typeof arg === 'function') {
        const id = functionArgs.push(arg) - 1;
        return {
          isFunction: true,
          id
        };
      }

      return arg;
    }); // This uses the testPath as the url so that if there are relative imports
    // in the inline code, the relative imports are resolved relative to the test file

    const url = `http://localhost:${port}/${testPath}?inline-code=${encodedCode}&build-id=${buildStatus.buildId}`;
    const outputHandle = await page.evaluateHandle(new Function('...args', `
          const argsWithFuncs = args.map(arg => {
            if (typeof arg === 'object' && arg?.isFunction) {
              return async (...args) => {
                return await window._pleasantestCallFunction(arg.id, args);
              }
            }
            return arg
          })
          window._pleasantestArgs = argsWithFuncs;
          return import(${JSON.stringify(url)})
            .then(async mod => {
              window._pleasantestArgs = undefined;
              return { success: true, result: { ...mod } }
            })
            .catch(e => ({
              success: false,
              result: e instanceof Error ? { message: e.message, stack: e.stack } : e
            }))`), ...(Array.isArray(argsWithFuncsAsObjs) ? argsWithFuncsAsObjs : []));
    const {
      success,
      result
    } = Object.fromEntries(await outputHandle.getProperties());
    const errorsFromBuild = buildStatus.complete(); // It only throws the first one but that is probably OK

    if (errorsFromBuild) throw errorsFromBuild[0];
    if (await success.jsonValue()) return Object.fromEntries(await result.getProperties());
    await sourceMapErrorFromBrowser(await result.jsonValue(), requestCache, port, runJS);
    throw await result.jsonValue();
  }, runJS);

  const injectHTML = (html, {
    executeScriptTags = true
  } = {}) => asyncHookTracker.addHook(() => page.evaluate((html, executeScriptTags) => {
    document.body.innerHTML = html;

    if (executeScriptTags) {
      // Scripts injected with innerHTML are not executed by default;
      // To get the browser to execute them we must manually create the script tags
      // and replace them
      const scripts = document.body.querySelectorAll('script');

      for (const script of scripts) {
        const newScript = document.createElement('script');
        newScript.text = script.innerHTML;

        for (const attribute of script.attributes) {
          newScript.setAttribute(attribute.name, attribute.value);
        }

        script.replaceWith(newScript);
      }
    }
  }, html, executeScriptTags), injectHTML);

  const injectCSS = css => asyncHookTracker.addHook(() => page.evaluate(css => {
    const styleTag = document.createElement('style');
    styleTag.innerHTML = css;
    document.head.append(styleTag);
  }, css), injectCSS);

  const loadCSS = cssPath => asyncHookTracker.addHook(async () => {
    const fullPath = path.isAbsolute(cssPath) ? path.relative(process.cwd(), cssPath) : path.join(path.dirname(testPath), cssPath);
    await page.evaluate(`import(${JSON.stringify(`http://localhost:${port}/${fullPath}?import`)})`);
  }, loadCSS);

  const loadJS = jsPath => asyncHookTracker.addHook(async () => {
    const fullPath = jsPath.startsWith('.') ? path.join(path.dirname(testPath), jsPath) : jsPath;
    const buildStatus = createBuildStatusTracker();
    const url = `http://localhost:${port}/${fullPath}?build-id=${buildStatus.buildId}`;
    const res = await page.evaluate(`import(${JSON.stringify(url)})
          .then(mod => {})
          .catch(e => e instanceof Error
            ? { message: e.message, stack: e.stack }
            : e)`);
    const errorsFromBuild = buildStatus.complete(); // It only throws the first one but that is probably OK

    if (errorsFromBuild) throw errorsFromBuild[0];
    await sourceMapErrorFromBrowser(res, requestCache, port, loadJS);
  }, loadJS);

  const utils = {
    runJS,
    injectCSS,
    injectHTML,
    loadCSS,
    loadJS
  };
  const screen = getQueriesForElement(page, asyncHookTracker); // The | null is so you can pass directly the result of page.$() which returns null if not found

  const within = element => {
    assertElementHandle(element, within);
    return getQueriesForElement(page, asyncHookTracker, element);
  };

  const waitFor$1 = (cb, opts = {}) => waitFor(page, asyncHookTracker, cb, opts, waitFor$1);

  return {
    screen,
    utils,
    page,
    within,
    waitFor: waitFor$1,
    user: await pleasantestUser(page, asyncHookTracker, { ...defaultOptions.user,
      ...userOpts
    }),
    asyncHookTracker,
    cleanupServer: () => closeServer()
  };
};

let defaultOptions = {};
const configureDefaults = options => {
  defaultOptions = options;
};
const devices = puppeteer.devices;
afterAll(async () => {
  await cleanupClientRuntimeServer();
});

exports.ElementHandle = puppeteer.ElementHandle;
exports.JSHandle = puppeteer.JSHandle;
exports.accessibilityTreeSnapshotSerializer = accessibilityTreeSnapshotSerializer;
exports.configureDefaults = configureDefaults;
exports.devices = devices;
exports.getAccessibilityTree = getAccessibilityTreeWrapper;
exports.makeCallableJSHandle = makeCallableJSHandle;
exports.withBrowser = withBrowser;
