import { promises } from 'node:fs';
import os from 'node:os';
import path__default from 'node:path';
import * as puppeteer from 'puppeteer';

process.on('message', async ({
  browser,
  headless
}) => {
  try {
    if (browser !== 'chromium') throw new Error(`unrecognized browser: ${browser}`);
    if (typeof headless !== 'boolean') throw new Error('headless must be a boolean');
    let userDataDir;

    if (browser === 'chromium') {
      userDataDir = await promises.mkdtemp(os.tmpdir() + path__default.sep);
      const prefs = JSON.stringify({
        devtools: {
          preferences: {
            'panel-selectedTab': JSON.stringify('console'),
            currentDockState: JSON.stringify('right'),
            'InspectorView.splitViewState': JSON.stringify({
              horizontal: {
                size: 600
              },
              vertical: {
                size: 450
              }
            })
          }
        }
      });
      const prefsPath = path__default.join(userDataDir, 'Default', 'Preferences');
      await promises.mkdir(path__default.dirname(prefsPath), {
        recursive: true
      });
      await promises.writeFile(prefsPath, prefs);
    }

    const browserInstance = await puppeteer.launch({
      headless,
      devtools: !headless,
      ignoreDefaultArgs: [// Don't pop up "Chrome is being controlled by automated software"
      '--enable-automation', // Unsupported flag that pops up a warning
      '--enable-blink-features=IdleDetection'],
      args: [// Don't pop up "Chrome is not your default browser"
      '--no-default-browser-check', '--disable-features=ChromeLabs' // Remove little beaker icon next to URL bar
      ],
      userDataDir
    });
    const allPages = await browserInstance.pages(); // Close startup page

    await Promise.all(allPages.map(p => p.close()));
    const browserWSEndpoint = browserInstance.wsEndpoint();
    process.send({
      browserWSEndpoint
    }); // eslint-disable-next-line @cloudfour/n/no-process-exit

    browserInstance.on('disconnected', () => process.exit());
  } catch (error) {
    process.send({
      error: error.message
    });
  }
});
